//******************************************************************************
//
//  Xilinx, Inc. 2002                 www.xilinx.com
//
//
//*******************************************************************************
//
//  File name :       monitor_ctrl.v
//
//  Description :     This module implements the real time window 
//                    monitoring control module.  
//                    
//  Date - revision : 11/30/2004
//
//  Author :          Tze Yi Yeoh
//
//  Contact : e-mail  hotline@xilinx.com
//            phone   + 1 800 255 7778 
//
//  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are 
//              provided to you "as is". Xilinx and its licensors make and you 
//              receive no warranties or conditions, express, implied, 
//              statutory or otherwise, and Xilinx specifically disclaims any 
//              implied warranties of merchantability, non-infringement, or 
//              fitness for a particular purpose. Xilinx does not warrant that 
//              the functions contained in these designs will meet your 
//              requirements, or that the operation of these designs will be 
//              uninterrupted or error free, or that defects in the Designs 
//              will be corrected. Furthermore, Xilinx does not warrant or 
//              make any representations regarding use or the results of the 
//              use of the designs in terms of correctness, accuracy, 
//              reliability, or otherwise. 
//
//              LIMITATION OF LIABILITY. In no event will Xilinx or its 
//              licensors be liable for any loss of data, lost profits, cost 
//              or procurement of substitute goods or services, or for any 
//              special, incidental, consequential, or indirect damages 
//              arising from the use or operation of the designs or 
//              accompanying documentation, however caused and on any theory 
//              of liability. This limitation will apply even if Xilinx 
//              has been advised of the possibility of such damage. This 
//              limitation shall apply not-withstanding the failure of the 
//              essential purpose of any limited remedies herein. 
//
//  Copyright  2002 Xilinx, Inc.
//  All rights reserved 
// 
//*****************************************************************************

`timescale 1ps/1ps

module monitor_ctrl (
   datain,
   datain_monitor,
   rst,
   monitor_en, 
   clkdiv,
   dlyce_window,
   dlyinc_window,
   dlyce,
   dlyinc,
   monitor_done
   );


input [3:0] datain;
input [3:0] datain_monitor;
input rst; 
input monitor_en;
input clkdiv;
output dlyce_window;
output dlyinc_window;
output dlyce;
output dlyinc;
output monitor_done;

reg flag_right;
reg flag_left;
reg dlyce_window;
reg dlyinc_window;
reg dlyce;
reg dlyinc;
reg monitor_done;
reg [7:0] counter;
reg cnt_rst;
reg cnt_inc;

// Set the WIDTH of the monitoring window.
// In this example, width of window = 6 taps (500 ps) for 1Gbps
parameter [6:0] WIDTH = 7'b0000110; 

// Window Monitoring Control State Machine
//Declare state machine parameters
parameter START            = 5'b00000; //0
parameter START_WAIT1      = 5'b00001; //1
parameter START_WAIT2      = 5'b00010; //2
parameter DEC              = 5'b00011; //3
parameter WAIT_DEC         = 5'b00100; //4
parameter CHECK_LEFT       = 5'b00101; //5
parameter LEFT_EDGE        = 5'b00110; //6
parameter INC              = 5'b00111; //7
parameter WAIT_INC         = 5'b01000; //8
parameter CHECK_RIGHT      = 5'b01001; //9
parameter RIGHT_EDGE       = 5'b01010; //10
parameter MOVE_RIGHT       = 5'b01011; //11
parameter WAIT_MOVE_RIGHT1 = 5'b01100; //12
parameter WAIT_MOVE_RIGHT2 = 5'b01101; //13
parameter MOVE_LEFT        = 5'b01110; //14
parameter WAIT_MOVE_LEFT1  = 5'b01111; //15
parameter WAIT_MOVE_LEFT2  = 5'b10000; //16
parameter DONE             = 5'b10001; //17

reg [4:0] Current_State = START;
reg [4:0] Next_State;

//

// All-purpose counter
always @ (posedge clkdiv or posedge rst) 
   if (rst)    
      counter <= 8'h00;
   else if (cnt_rst)    
      counter <= 8'h00;
   else if (cnt_inc)
      counter <= counter + 1'b1; 

// Current State Logic
always @ (posedge clkdiv or posedge rst) 
   if (rst)    
      Current_State <= START ;
   else if (monitor_en)
      Current_State <= Next_State;

// Output forming logic
always @ *
begin
   case (Current_State)
      START: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b1;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end  
		
      START_WAIT1: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end

      START_WAIT2: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end
		
	 // Decrement tap delay chain by 1 tap
      DEC: begin
         dlyce_window = 1'b1;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b1;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end   

      WAIT_DEC: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end

      CHECK_LEFT: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end
		
      // Check to see if left edge of monitoring window reached
      LEFT_EDGE: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b1;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         monitor_done = 1'b0;
			
         // If the left edge of the monitoring window is reached, 
         // compare datain from Data ISERDES and datain from 
         // Monitor ISERDES. If datain Data ISERDES does not match datain 
         // Monitor ISERDES, increment tap delay chain for both Data ISERDES
         // and Monitor ISERDES.
         if (datain != ~datain_monitor) // datain_monitor = ~datain due to IBUFDS_DIFF_OUT
            flag_left = 1'b1;
      end          

      // Increment tap delay chain by 1 tap
      INC: begin
         dlyce_window = 1'b1;   
         dlyinc_window = 1'b1;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b1;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;							     	         
      end   

      WAIT_INC: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end

      CHECK_RIGHT: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end
		
      // Check to see if right edge of monitoring window reached
      RIGHT_EDGE: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b1;
         cnt_inc =1'b0;
         flag_left = 1'b0;
         monitor_done = 1'b0;
         // If the right edge of the monitoring window is reached, 
         // compare datain from Data ISERDES and datain from 
         // Monitor ISERDES. If datain Data ISERDES does not match datain 
         // Monitor ISERDES, decrement tap delay chain for both Data ISERDES
         // and Monitor ISERDES.
         if (datain != ~datain_monitor) // datain_monitor = ~datain due to IBUFDS_DIFF_OUT
            flag_right = 1'b1;				      
      end 

	 // Increment the tap delay chain for both the Data ISERDES and 
	 // Monitor ISERDES when data falls outside the left edge of the 
	 // data valid window.
      MOVE_RIGHT: begin 
         dlyce_window = 1'b1;   
         dlyinc_window = 1'b1;
         dlyce = 1'b1;   
         dlyinc = 1'b1;  
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;	
    	 end

      WAIT_MOVE_RIGHT1: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end

      WAIT_MOVE_RIGHT2: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end
		
	 // Decrement the tap delay chain for both the Data ISERDES and 
	 // Monitor ISERDES when data falls outside the right edge of the 
	 // data valid window.
      MOVE_LEFT: begin        
         dlyce_window = 1'b1; 
         dlyinc_window = 1'b0;
         dlyce = 1'b1;  
         dlyinc = 1'b0;  
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;	
      end

      WAIT_MOVE_LEFT1: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end

      WAIT_MOVE_LEFT2: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end
		
      DONE: begin 
         dlyce_window = 1'b0; 
         dlyinc_window = 1'b0;
         dlyce = 1'b0;  
         dlyinc = 1'b0;  
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b1;
      end

      default: begin
         dlyce_window = 1'b0;
         dlyinc_window = 1'b0;
         dlyce = 1'b0;
         dlyinc = 1'b0;
         cnt_rst = 1'b0;
         cnt_inc =1'b0;
         flag_right = 1'b0;
         flag_left = 1'b0;  
         monitor_done = 1'b0;
      end
   endcase
end

always @ *
   case (Current_State)
      START:
	       Next_State <= START_WAIT1;

      START_WAIT1: 
         Next_State = START_WAIT2;			     	         
       
      START_WAIT2: 
	      Next_State = DEC;			     	         
       
	 // Decrement tap delay chain by 1 tap
      DEC: 
	      Next_State = WAIT_DEC;			     	         
         
      WAIT_DEC: 
	      Next_State = CHECK_LEFT;			     	         
        
      // Check to see if left edge of monitoring window reached
      CHECK_LEFT: 
      // If left edge of the monitoring window reached
         if (counter == WIDTH)	    
			   // If the left edge of the monitoring window is reached, 
			   // compare datain from Data ISERDES and datain from 
			   // Monitor ISERDES. If datain Data ISERDES does not match datain 
			   // Monitor ISERDES, increment tap delay chain for both Data ISERDES
			   // and Monitor ISERDES.
			   Next_State = LEFT_EDGE;    
         // If left edge of the monitoring window not reached, keep
         // decrementing tap delay chain of Monitor ISERDES.
         else
            Next_State = DEC; 	         

      LEFT_EDGE: 
			// If the left edge of the monitoring window is reached, 
			// compare datain from Data ISERDES and datain from 
			// Monitor ISERDES. If datain Data ISERDES does not match datain 
			// Monitor ISERDES, increment tap delay chain for both Data ISERDES
			// and Monitor ISERDES.
         if (datain == ~datain_monitor) // datain_monitor = ~datain due to IBUFDS_DIFF_OUT
            Next_State = INC;
         else	
         // If the left edge of the monitoring window is reached, 
         // compare datain from Data ISERDES and datain from 
         // Monitor ISERDES. If datain Data ISERDES matches datain 
         // Monitor ISERDES, increment tap delay chain for Monitor ISERDES
         // and check the right edge of the monitoring window.							      
            Next_State = MOVE_RIGHT;	    	
			         
      // Increment tap delay chain by 1 tap
      INC: 
         Next_State = WAIT_INC;			     	         
         
      WAIT_INC: 
         Next_State = CHECK_RIGHT;			     	         
       
      // Check to see if right edge of monitoring window reached
      CHECK_RIGHT: 
         // If left edge of the monitoring window reached
         if (counter == WIDTH)	    
         // If the right edge of the monitoring window is reached, 
         // compare datain from Data ISERDES and datain from 
         // Monitor ISERDES. If datain Data ISERDES does not match datain 
         // Monitor ISERDES, decrement tap delay chain for both Data ISERDES
         // and Monitor ISERDES.
            Next_State = RIGHT_EDGE;				       
         else
         // If right edge of the monitoring window not reached, keep
         // incrementing tap delay chain of Monitor ISERDES.		   
            Next_State = INC; 	         
       
      RIGHT_EDGE:  
         if (datain == ~datain_monitor) // datain_monitor = ~datain due to IBUFDS_DIFF_OUT
            Next_State = DONE;
         else	
         // If the right edge of the monitoring window is reached, 
         // compare datain from Data ISERDES and datain from 
         // Monitor ISERDES. If datain Data ISERDES matches datain 
         // Monitor ISERDES, decrement tap delay chain for Monitor ISERDES
         // and check the left edge of the monitoring window.							      
            Next_State = MOVE_LEFT;	

      // Increment the tap delay chain for both the Data ISERDES and 
      // Monitor ISERDES when data falls outside the left edge of the 
      // data valid window.
      MOVE_RIGHT:  
         Next_State = WAIT_MOVE_RIGHT1 ;
   	 
      WAIT_MOVE_RIGHT1:  
         Next_State = WAIT_MOVE_RIGHT2 ;

      WAIT_MOVE_RIGHT2:  
         Next_State = LEFT_EDGE ;
			   	
     // Decrement the tap delay chain for both the Data ISERDES and 
     // Monitor ISERDES when data falls outside the right edge of the 
     // data valid window.
     MOVE_LEFT:  
        Next_State = WAIT_MOVE_LEFT1 ;
   	
     WAIT_MOVE_LEFT1:  
        Next_State = WAIT_MOVE_LEFT2;

     WAIT_MOVE_LEFT2:  
        Next_State = RIGHT_EDGE;
		     	
     DONE:  
        Next_State = START;
   	
     default: Next_State = START ;
   endcase
endmodule

