//******************************************************************************
//
//  Xilinx, Inc. 2002                 www.xilinx.com
//
//
//*******************************************************************************
//
//  File name :       testb_16ch_training_monitor.v
//
//  Description :     This module implements the testbench for the DPA design
// 
//                    
//  Date - revision : 11/30/2004
//
//  Author :          Tze Yi Yeoh
//
//  Contact : e-mail  hotline@xilinx.com
//            phone   + 1 800 255 7778 
//
//  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are 
//              provided to you "as is". Xilinx and its licensors make and you 
//              receive no warranties or conditions, express, implied, 
//              statutory or otherwise, and Xilinx specifically disclaims any 
//              implied warranties of merchantability, non-infringement, or 
//              fitness for a particular purpose. Xilinx does not warrant that 
//              the functions contained in these designs will meet your 
//              requirements, or that the operation of these designs will be 
//              uninterrupted or error free, or that defects in the Designs 
//              will be corrected. Furthermore, Xilinx does not warrant or 
//              make any representations regarding use or the results of the 
//              use of the designs in terms of correctness, accuracy, 
//              reliability, or otherwise. 
//
//              LIMITATION OF LIABILITY. In no event will Xilinx or its 
//              licensors be liable for any loss of data, lost profits, cost 
//              or procurement of substitute goods or services, or for any 
//              special, incidental, consequential, or indirect damages 
//              arising from the use or operation of the designs or 
//              accompanying documentation, however caused and on any theory 
//              of liability. This limitation will apply even if Xilinx 
//              has been advised of the possibility of such damage. This 
//              limitation shall apply not-withstanding the failure of the 
//              essential purpose of any limited remedies herein. 
//
//  Copyright  2002 Xilinx, Inc.
//  All rights reserved 
// 
//*****************************************************************************
`timescale 1ps/1ps

module testb_16ch_training_monitor();

parameter num_chan = 16;
 
wire [3:0] dataout; 
wire done;
wire ready;
reg train_done;
reg clkin;
reg clkdlyctrl;
reg [num_chan-1:0] dip;  
wire [num_chan-1:0] din; 
reg txdata;
reg dlyrst;
reg dlyctrlrst;
reg rst;
reg train_en;
reg [19:0] test;
reg [19:0] realdata;
integer counter;
real R;
parameter [19:0] TEST_VECTOR = 20'b00000000001111111111; 
parameter [19:0] REAL_DATA = 20'b01010101010101010101;
parameter tjmax = 300;

// Instantiate top level module
top_16ch_training_monitor dut ( 
   .dataout(dataout),
   .train_done(done),
   .ready(ready),
   .clkin(clkin),
   .clkdlyctrl(clkdlyctrl),
   .dip(dip),
   .din(din),
   .dlyrst(dlyrst),
   .dlyctrlrst(dlyctrlrst),
   .rst(rst),
   .train_en(train_en)
   );

assign din = ~dip;
assign glbl.GSR = rst;

// Input forwarded source synchronous clock
initial
begin
   clkin = 1'b0;
   forever clkin = #1250 ~clkin; 
end

	// 200 Mhz REFCLK for IDELAYCTRL module
initial
begin
   clkdlyctrl = 1'b0;
   forever clkdlyctrl = #2500 ~clkdlyctrl;
end

initial
begin
   test = TEST_VECTOR;
   realdata = REAL_DATA;
   rst = 1'b1;
   train_en = 1'b0;
   dlyrst = 1'b0;
   dlyctrlrst = 1'b0;
   #100
   rst = 1'b1;
   dlyrst = 1'b1;
   dlyctrlrst = 1'b1;
   #1000
   rst = 1'b0;
   dlyrst = 1'b0;
   dlyctrlrst = 1'b0;
end

// Generate input serial data stream
always @ (posedge clkin or negedge clkin or posedge rst or posedge train_done)
   if (rst)
      txdata  <= 1'b0;     
   else
   // At reset, send training pattern 
      if (~train_done)
      begin
//         txdata <= #(R*tjmax + 900) test[19];
         txdata <= test[19];
         test <= {test[18:0] , test[19]};
      end
      else
      // After training is done, send real data (oscillating signal) 
      begin		    
         txdata <= realdata[19];
         realdata <= {realdata[18:0] , realdata[19]};
      end

generate
genvar i;
for (i=0; i<=13; i=i+1)
begin : generate_dip
   always @ (posedge clkin or negedge clkin or posedge rst)
      if (rst)
         dip[i]  <= 1'b0;
      else
         dip[i] <= #11500 txdata;  
end
endgenerate

// Skew Channel 1 by -1 bit period
always @ (posedge clkin or negedge clkin or posedge rst)
   if (rst)
      dip[14]  <= 1'b0;
   else
      dip[14] <= #10500 txdata;  

// Skew Channel 0 by +1 bit period
always @ (posedge clkin or negedge clkin or posedge rst)
   if (rst)
      dip[15]  <= 1'b0;
	else
      dip[15] <= #9500 txdata;  
   
always @(posedge done or posedge rst)
   if (done)
      train_done = 1'b1;
   else 
      train_done = 1'b0;

always @(posedge done or negedge rst)
   if (done)
      train_en = 1'b0;
   else if (!rst) 
	begin
      #20000
      train_en = 1'b1;
	end
	else
	   train_en = train_en;

endmodule