//******************************************************************************
//
//  Xilinx, Inc. 2002                 www.xilinx.com
//
//
//*******************************************************************************
//
//  File name :       top_16ch_training_monitor.v
//
//  Description :     This module is the top level module for the DPA design
// 
//                    
//  Date - revision : 11/30/2004
//
//  Author :          Tze Yi Yeoh
//
//  Contact : e-mail  hotline@xilinx.com
//            phone   + 1 800 255 7778 
//
//  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are 
//              provided to you "as is". Xilinx and its licensors make and you 
//              receive no warranties or conditions, express, implied, 
//              statutory or otherwise, and Xilinx specifically disclaims any 
//              implied warranties of merchantability, non-infringement, or 
//              fitness for a particular purpose. Xilinx does not warrant that 
//              the functions contained in these designs will meet your 
//              requirements, or that the operation of these designs will be 
//              uninterrupted or error free, or that defects in the Designs 
//              will be corrected. Furthermore, Xilinx does not warrant or 
//              make any representations regarding use or the results of the 
//              use of the designs in terms of correctness, accuracy, 
//              reliability, or otherwise. 
//
//              LIMITATION OF LIABILITY. In no event will Xilinx or its 
//              licensors be liable for any loss of data, lost profits, cost 
//              or procurement of substitute goods or services, or for any 
//              special, incidental, consequential, or indirect damages 
//              arising from the use or operation of the designs or 
//              accompanying documentation, however caused and on any theory 
//              of liability. This limitation will apply even if Xilinx 
//              has been advised of the possibility of such damage. This 
//              limitation shall apply not-withstanding the failure of the 
//              essential purpose of any limited remedies herein. 
//
//  Copyright  2002 Xilinx, Inc.
//  All rights reserved 
// 
//*****************************************************************************

`timescale 1ps/1ps

module top_16ch_training_monitor (
	 dataout,
	 train_done,
	 ready,
	 clkin,
	 clkdlyctrl,
	 dip,
      din,
	 dlyrst,
	 dlyctrlrst,
	 rst,
	 train_en
	);

// Parameter that determines the number of channels instantiated.
parameter num_chan = 16;

output [3:0] dataout;
output train_done;
output ready;
input  clkin;
input  clkdlyctrl;
input  [num_chan-1:0] dip;
input  [num_chan-1:0] din;
input  dlyrst;
input  dlyctrlrst;
input  rst;
input  train_en;

wire [num_chan-1:0] data_in;
wire [num_chan-1:0] window_in;
reg [num_chan-1:0] chan_sel;
reg [num_chan-1:0] chan_sel1;
reg [num_chan-1:0] dataout_sel;
reg [num_chan-1:0] dataout_sel_int;
reg [1:0]  polarity;
wire deskew_done;
wire bitslip_done;
wire monitor_done;
wire bitslip_error;
reg  train_done;
wire [num_chan-1:0] dlyce;
wire [num_chan-1:0] dlyinc;
wire [num_chan-1:0] dlyce_iserdes_window;
wire [num_chan-1:0] dlyinc_iserdes_window;
wire dlyce_training;
wire dlyinc_training;
wire dlyce_window;
wire dlyinc_window;
wire dlyce_iserdes_window1;
wire dlyinc_iserdes_window1;
wire dlyce1;
wire dlyinc1;
wire [num_chan-1:0] bitslip;
wire bitslip_training;
wire iobclk;
wire iobclk_ibuf;
wire clkdiv;
wire refclk;
wire refclk_ibufg;
wire [num_chan-1:0] q1;
wire [num_chan-1:0] q2;
wire [num_chan-1:0] q3;
wire [num_chan-1:0] q4;
wire [num_chan-1:0] w1;
wire [num_chan-1:0] w2;
wire [num_chan-1:0] w3;
wire [num_chan-1:0] w4;
wire [3:0] datain;
wire [3:0] datain_monitor;
wire [3:0] dataout15;
wire [3:0] dataout14;
wire [3:0] dataout13;
wire [3:0] dataout12;
wire [3:0] dataout11;
wire [3:0] dataout10;
wire [3:0] dataout9;
wire [3:0] dataout8;
wire [3:0] dataout7;
wire [3:0] dataout6;
wire [3:0] dataout5;
wire [3:0] dataout4;
wire [3:0] dataout3;
wire [3:0] dataout2;
wire [3:0] dataout1;
wire [3:0] dataout0;
reg [3:0] dataout15_R1;
reg [3:0] dataout14_R1;
reg [3:0] dataout13_R1;
reg [3:0] dataout12_R1;
reg [3:0] dataout11_R1;
reg [3:0] dataout10_R1;
reg [3:0] dataout9_R1;
reg [3:0] dataout8_R1;
reg [3:0] dataout7_R1;
reg [3:0] dataout6_R1;
reg [3:0] dataout5_R1;
reg [3:0] dataout4_R1;
reg [3:0] dataout3_R1;
reg [3:0] dataout2_R1;
reg [3:0] dataout1_R1;
reg [3:0] dataout0_R1;
wire [3:0] dataout_int15;
wire [3:0] dataout_int14;
wire [3:0] dataout_int13;
wire [3:0] dataout_int12;
wire [3:0] dataout_int11;
wire [3:0] dataout_int10;
wire [3:0] dataout_int9;
wire [3:0] dataout_int8;
wire [3:0] dataout_int7;
wire [3:0] dataout_int6;
wire [3:0] dataout_int5;
wire [3:0] dataout_int4;
wire [3:0] dataout_int3;
wire [3:0] dataout_int2;
wire [3:0] dataout_int1;
wire [3:0] dataout_int0;
wire [3:0] windowout15;
wire [3:0] windowout14;
wire [3:0] windowout13;
wire [3:0] windowout12;
wire [3:0] windowout11;
wire [3:0] windowout10;
wire [3:0] windowout9;
wire [3:0] windowout8;
wire [3:0] windowout7;
wire [3:0] windowout6;
wire [3:0] windowout5;
wire [3:0] windowout4;
wire [3:0] windowout3;
wire [3:0] windowout2;
wire [3:0] windowout1;
wire [3:0] windowout0;

reg [3:0] windowout15_R1;
reg [3:0] windowout14_R1;
reg [3:0] windowout13_R1;
reg [3:0] windowout12_R1;
reg [3:0] windowout11_R1;
reg [3:0] windowout10_R1;
reg [3:0] windowout9_R1;
reg [3:0] windowout8_R1;
reg [3:0] windowout7_R1;
reg [3:0] windowout6_R1;
reg [3:0] windowout5_R1;
reg [3:0] windowout4_R1;
reg [3:0] windowout3_R1;
reg [3:0] windowout2_R1;
reg [3:0] windowout1_R1;
reg [3:0] windowout0_R1;

// Dataout Channel Select Mux
wire [3:0] muxout_d0d1;
wire [3:0] muxout_d2d3;
wire [3:0] muxout_d4d5;
wire [3:0] muxout_d6d7;
wire [3:0] muxout_d8d9;
wire [3:0] muxout_d10d11;
wire [3:0] muxout_d12d13;
wire [3:0] muxout_d14d15;

wire [3:0] muxout_d0_to_d3;
wire [3:0] muxout_d4_to_d7;
wire [3:0] muxout_d8_to_d11;
wire [3:0] muxout_d12_to_d15;

reg [3:0] muxout_d0_to_d3_R1;
reg [3:0] muxout_d4_to_d7_R1;
reg [3:0] muxout_d8_to_d11_R1;
reg [3:0] muxout_d12_to_d15_R1;

wire [3:0] muxout_d0_to_d7;
wire [3:0] muxout_d8_to_d15;

// Window Monitor Channel Select Mux
wire [3:0] muxout_w0w1;
wire [3:0] muxout_w2w3;
wire [3:0] muxout_w4w5;
wire [3:0] muxout_w6w7;
wire [3:0] muxout_w8w9;
wire [3:0] muxout_w10w11;
wire [3:0] muxout_w12w13;
wire [3:0] muxout_w14w15;

wire [3:0] muxout_w0_to_w3;
wire [3:0] muxout_w4_to_w7;
wire [3:0] muxout_w8_to_w11;
wire [3:0] muxout_w12_to_w15;

reg [3:0] muxout_w0_to_w3_R1;
reg [3:0] muxout_w4_to_w7_R1;
reg [3:0] muxout_w8_to_w11_R1;
reg [3:0] muxout_w12_to_w15_R1;

wire [3:0] muxout_w0_to_w7;
wire [3:0] muxout_w8_to_w15;

wire [2:0] edgei_training;
reg [7:0] counter;
reg cnt_rst;
reg [3:0] op_sel;

parameter [num_chan-1:0]
  CHAN0  = 16'b0000000000000001,
  CHAN1  = 16'b0000000000000010,
  CHAN2  = 16'b0000000000000100,
  CHAN3  = 16'b0000000000001000,
  CHAN4  = 16'b0000000000010000,
  CHAN5  = 16'b0000000000100000,
  CHAN6  = 16'b0000000001000000,
  CHAN7  = 16'b0000000010000000,
  CHAN8  = 16'b0000000100000000,
  CHAN9  = 16'b0000001000000000,
  CHAN10 = 16'b0000010000000000,
  CHAN11 = 16'b0000100000000000,
  CHAN12 = 16'b0001000000000000,
  CHAN13 = 16'b0010000000000000,
  CHAN14 = 16'b0100000000000000,
  CHAN15 = 16'b1000000000000000;

// Instantiate IDELAYCTRL module because the ISERDES is used in 
// VARIABLE tap delay chain mode. 
IDELAYCTRL dlyctrl (
   .RDY(ready),
   .REFCLK(refclk),
   .RST(dlyctrlrst)
   );

IBUFG ibufg_module (
   .O(refclk_ibufg),
   .I(clkdlyctrl)
   );

BUFG bufg_module (
   .O(refclk),
   .I(refclk_ibufg)
   );

IBUF ibuf_module (
   .O(iobclk_ibuf),
   .I(clkin)
   );

// Instantiate regional clock buffers. One regional clock buffer
// can accomodate up to 3 clock regions (96 IOBs). 
BUFIO bufio_module (
   .O(iobclk),
   .I(iobclk_ibuf)
   );

// Note that the clock divide is set to 2 not 4 for a 1:4 
// deserialization factor in DDR mode.
BUFR #(
   .BUFR_DIVIDE("2")
	)
bufr_module (
   .O(clkdiv),
   .CE(1'b1),
   .CLR(1'b0),
   .I(iobclk)
   );     
// synthesis attribute buffer_type of clkdiv is none

// Instantiate data differential input buffers
IBUFDS_DIFF_OUT datain_buf[num_chan-1:0] (
   .I(dip[num_chan-1:0]),
   .IB(din[num_chan-1:0]),
   .O(data_in[num_chan-1:0]),
   .OB(window_in[num_chan-1:0])
	);

// Instantiate the master data channel ISERDES
ISERDES #(
   .BITSLIP_ENABLE("TRUE"),
   .DATA_RATE("DDR"),
   .DATA_WIDTH(4),
   .INIT_Q1(1'b0),
   .INIT_Q2(1'b0),
   .INIT_Q3(1'b0),
   .INIT_Q4(1'b0),
   .INTERFACE_TYPE("NETWORKING"),
   .IOBDELAY("IFD"),
   .IOBDELAY_TYPE("VARIABLE"),
   .IOBDELAY_VALUE(0),
   .NUM_CE(1),
   .SERDES_MODE("MASTER"),
   .SRVAL_Q1(1'b0),
   .SRVAL_Q2(1'b0),
   .SRVAL_Q3(1'b0),
   .SRVAL_Q4(1'b0) 
	)
data_chan_master_[num_chan-1:0] (
   .O(), 
   .Q1(q1[num_chan-1:0]), 
   .Q2(q2[num_chan-1:0]), 
   .Q3(q3[num_chan-1:0]), 
   .Q4(q4[num_chan-1:0]), 
   .Q5(),
   .Q6(),
   .SHIFTOUT1(), 
   .SHIFTOUT2(),
   .BITSLIP(bitslip[num_chan-1:0]),
   .CE1(1'b1), 
   .CE2(1'b1),
   .CLK(iobclk), 
   .CLKDIV(clkdiv), 				   
   .D(data_in[num_chan-1:0]), 
   .DLYCE(dlyce[num_chan-1:0]),
   .DLYINC(dlyinc[num_chan-1:0]),
   .DLYRST(dlyrst),
   .OCLK(1'b0), 
   .REV(1'b0),
   .SHIFTIN1(1'b0), 
   .SHIFTIN2(1'b0),
   .SR(rst)		  
   );

// Instantiate the window monitor data channel ISERDES 
ISERDES #(
   .BITSLIP_ENABLE("TRUE"),
   .DATA_RATE("DDR"),
   .DATA_WIDTH(4),
   .INIT_Q1(1'b0),
   .INIT_Q2(1'b0),
   .INIT_Q3(1'b0),
   .INIT_Q4(1'b0),
   .INTERFACE_TYPE("NETWORKING"),
   .IOBDELAY("IFD"),
   .IOBDELAY_TYPE("VARIABLE"),
   .IOBDELAY_VALUE(3),
   .NUM_CE(1),
   .SERDES_MODE("MASTER"),
   .SRVAL_Q1(1'b0),
   .SRVAL_Q2(1'b0),
   .SRVAL_Q3(1'b0),
   .SRVAL_Q4(1'b0) 
   )
window_monitor_[num_chan-1:0] (
   .O(), 
   .Q1(w1[num_chan-1:0]), 
   .Q2(w2[num_chan-1:0]), 
   .Q3(w3[num_chan-1:0]), 
   .Q4(w4[num_chan-1:0]), 
   .Q5(),
   .Q6(),
   .SHIFTOUT1(), 
   .SHIFTOUT2(),
   .BITSLIP(bitslip[num_chan-1:0]),
   .CE1(1'b1), 
   .CE2(1'b1),
   .CLK(iobclk), 
   .CLKDIV(clkdiv), 				   
   .D(window_in[num_chan-1:0]), 
   .DLYCE(dlyce_iserdes_window[num_chan-1:0]),
   .DLYINC(dlyinc_iserdes_window[num_chan-1:0]),
   .DLYRST(dlyrst),
   .OCLK(1'b0), 
   .REV(1'b0),
   .SHIFTIN1(1'b0), 
   .SHIFTIN2(1'b0),
   .SR(rst)			  
   );

assign dataout15 = {{q4[15],q3[15]},{q2[15],q1[15]}};
assign dataout14 = {{q4[14],q3[14]},{q2[14],q1[14]}};
assign dataout13 = {{q4[13],q3[13]},{q2[13],q1[13]}};
assign dataout12 = {{q4[12],q3[12]},{q2[12],q1[12]}};
assign dataout11 = {{q4[11],q3[11]},{q2[11],q1[11]}};
assign dataout10 = {{q4[10],q3[10]},{q2[10],q1[10]}};
assign dataout9 = {{q4[9],q3[9]},{q2[9],q1[9]}};
assign dataout8 = {{q4[8],q3[8]},{q2[8],q1[8]}};
assign dataout7 = {{q4[7],q3[7]},{q2[7],q1[7]}};
assign dataout6 = {{q4[6],q3[6]},{q2[6],q1[6]}};
assign dataout5 = {{q4[5],q3[5]},{q2[5],q1[5]}};
assign dataout4 = {{q4[4],q3[4]},{q2[4],q1[4]}};
assign dataout3 = {{q4[3],q3[3]},{q2[3],q1[3]}};
assign dataout2 = {{q4[2],q3[2]},{q2[2],q1[2]}};
assign dataout1 = {{q4[1],q3[1]},{q2[1],q1[1]}};
assign dataout0 = {{q4[0],q3[0]},{q2[0],q1[0]}};

assign windowout15 = {{w4[15],w3[15]},{w2[15],w1[15]}};
assign windowout14 = {{w4[14],w3[14]},{w2[14],w1[14]}};
assign windowout13 = {{w4[13],w3[13]},{w2[13],w1[13]}};
assign windowout12 = {{w4[12],w3[12]},{w2[12],w1[12]}};
assign windowout11 = {{w4[11],w3[11]},{w2[11],w1[11]}};
assign windowout10 = {{w4[10],w3[10]},{w2[10],w1[10]}};
assign windowout9 = {{w4[9],w3[9]},{w2[9],w1[9]}};
assign windowout8 = {{w4[8],w3[8]},{w2[8],w1[8]}};
assign windowout7 = {{w4[7],w3[7]},{w2[7],w1[7]}};
assign windowout6 = {{w4[6],w3[6]},{w2[6],w1[6]}};
assign windowout5 = {{w4[5],w3[5]},{w2[5],w1[5]}};
assign windowout4 = {{w4[4],w3[4]},{w2[4],w1[4]}};
assign windowout3 = {{w4[3],w3[3]},{w2[3],w1[3]}};
assign windowout2 = {{w4[2],w3[2]},{w2[2],w1[2]}};
assign windowout1 = {{w4[1],w3[1]},{w2[1],w1[1]}};
assign windowout0 = {{w4[0],w3[0]},{w2[0],w1[0]}};

// Dummy OR function to OR all data outputs together. Replace with user logic													
assign dataout = (dataout_int15 | dataout_int14 | dataout_int13 | 
                  dataout_int12 | dataout_int11 | dataout_int10 | dataout_int9 |
                  dataout_int8 | dataout_int7 | dataout_int6 | dataout_int5 | 
                  dataout_int4 | dataout_int3 | dataout_int2 | dataout_int1 | dataout_int0); 

// Pipeline ISERDES outputs
always @(posedge rst or posedge clkdiv)
begin
   if (rst)
   begin	  
	 dataout15_R1 <= 1'h0;
	 dataout14_R1 <= 1'h0;
	 dataout13_R1 <= 1'h0;
	 dataout12_R1 <= 1'h0;
	 dataout11_R1 <= 1'h0;
	 dataout10_R1 <= 1'h0;
	 dataout9_R1 <= 1'h0;
	 dataout8_R1 <= 1'h0;
	 dataout7_R1 <= 1'h0;
	 dataout6_R1 <= 1'h0;
	 dataout5_R1 <= 1'h0;
	 dataout4_R1 <= 1'h0;
	 dataout3_R1 <= 1'h0;
	 dataout2_R1 <= 1'h0;
	 dataout1_R1 <= 1'h0;
	 dataout0_R1 <= 1'h0;

	 windowout15_R1 <= 1'h0;
	 windowout14_R1 <= 1'h0;
	 windowout13_R1 <= 1'h0;
	 windowout12_R1 <= 1'h0;
	 windowout11_R1 <= 1'h0;
	 windowout10_R1 <= 1'h0;
	 windowout9_R1 <= 1'h0;
	 windowout8_R1 <= 1'h0;
	 windowout7_R1 <= 1'h0;
	 windowout6_R1 <= 1'h0;
	 windowout5_R1 <= 1'h0;
	 windowout4_R1 <= 1'h0;
	 windowout3_R1 <= 1'h0;
	 windowout2_R1 <= 1'h0;
	 windowout1_R1 <= 1'h0;
	 windowout0_R1 <= 1'h0;

   end
   else
   begin
	 dataout15_R1 <= dataout15;
	 dataout14_R1 <= dataout14;
	 dataout13_R1 <= dataout13;
	 dataout12_R1 <= dataout12;
	 dataout11_R1 <= dataout11;
	 dataout10_R1 <= dataout10;
	 dataout9_R1 <= dataout9;
	 dataout8_R1 <= dataout8;
	 dataout7_R1 <= dataout7;
	 dataout6_R1 <= dataout6;
	 dataout5_R1 <= dataout5;
	 dataout4_R1 <= dataout4;
	 dataout3_R1 <= dataout3;
	 dataout2_R1 <= dataout2;
	 dataout1_R1 <= dataout1;
	 dataout0_R1 <= dataout0;
 
	 windowout15_R1 <= windowout15;
	 windowout14_R1 <= windowout14;
	 windowout13_R1 <= windowout13;
	 windowout12_R1 <= windowout12;
	 windowout11_R1 <= windowout11;
	 windowout10_R1 <= windowout10;
	 windowout9_R1 <= windowout9;
	 windowout8_R1 <= windowout8;
	 windowout7_R1 <= windowout7;
	 windowout6_R1 <= windowout6;
	 windowout5_R1 <= windowout5;
	 windowout4_R1 <= windowout4;
	 windowout3_R1 <= windowout3;
	 windowout2_R1 <= windowout2;
	 windowout1_R1 <= windowout1;
	 windowout0_R1 <= windowout0;

   end	    	              
end

// Bitslip adjust Mux that selects between ISERDES and pipeline outputs
assign dataout_int15 = dataout_sel[15] ? dataout15_R1 : dataout15; 
assign dataout_int14 = dataout_sel[14] ? dataout14_R1 : dataout14; 
assign dataout_int13 = dataout_sel[13] ? dataout13_R1 : dataout13; 
assign dataout_int12 = dataout_sel[12] ? dataout12_R1 : dataout12; 
assign dataout_int11 = dataout_sel[11] ? dataout11_R1 : dataout11; 
assign dataout_int10 = dataout_sel[10] ? dataout10_R1 : dataout10; 
assign dataout_int9 = dataout_sel[9] ? dataout9_R1 : dataout9; 
assign dataout_int8 = dataout_sel[8] ? dataout8_R1 : dataout8; 
assign dataout_int7 = dataout_sel[7] ? dataout7_R1 : dataout7; 
assign dataout_int6 = dataout_sel[6] ? dataout6_R1 : dataout6; 
assign dataout_int5 = dataout_sel[5] ? dataout5_R1 : dataout5; 
assign dataout_int4 = dataout_sel[4] ? dataout4_R1 : dataout4; 
assign dataout_int3 = dataout_sel[3] ? dataout3_R1 : dataout3; 
assign dataout_int2 = dataout_sel[2] ? dataout2_R1 : dataout2; 
assign dataout_int1 = dataout_sel[1] ? dataout1_R1 : dataout1; 
assign dataout_int0 = dataout_sel[0] ? dataout0_R1 : dataout0; 

// Instantiate deskew module that implements bit alignment algorithm
deskew deskew_module (
   .datain(datain),
   .edgei(edgei_training),
   .rst(rst),
   .deskew_en(op_sel[0]), 
   .clkdiv(clkdiv),
   .dlyce(dlyce_training),
   .dlyinc(dlyinc_training),
   .done(deskew_done)	  
   ); 
    
// Instantiate the control module that implements the word alignment algorithm
bitslip_ctrl bitslip_control_module (
   .datain(datain),
   .rst(rst),
   .bitslip_en(op_sel[1]), 
   .clkdiv(clkdiv),
   .bitslip(bitslip_training),
   .done(bitslip_done),
   .error(bitslip_error)	  
	); 

// Instantiate the control module that implements the real-time window monitoring algorithm
   monitor_ctrl monitor_ctrl_module (      
   .datain(datain), 
   .datain_monitor(datain_monitor), 
   .rst(rst),
   .monitor_en(train_done), 
   .clkdiv(clkdiv),
   .dlyce_window(dlyce_window),
   .dlyinc_window(dlyinc_window),
   .dlyce(dlyce_monitor),
   .dlyinc(dlyinc_monitor),
   .monitor_done(monitor_done)	  
   );

// Multi-purpose Counter
always @(posedge cnt_rst or posedge clkdiv)
begin
   if (cnt_rst)
      counter <= 8'h00;
   else if (counter[2])
      counter <= 8'h00;
   else
      counter <= counter + 1'b1;
end      

// Module that controls the operation select (bit align, word align, bitslip adjust
// or window monitoring) and channel select.
always @(posedge rst or posedge clkdiv)
begin
   if (rst)
   begin	  
      op_sel <= 4'b0000;
	 train_done <= 1'b0;
      chan_sel <= 16'b0000_0000_0000_0000;
	 cnt_rst <= 1'b1;
	 dataout_sel <= 16'b0000_0000_0000_0000;
   end
   else if (train_en)
   begin
      if (~|op_sel)
      begin	  
         op_sel <= 4'b0001;
	    train_done <= 1'b0;
         chan_sel <= 16'b0000_0000_0000_0001;
	    cnt_rst <= 1'b1;
	    dataout_sel <= 16'b0000_0000_0000_0000;
      end 
      else if (op_sel[0])
      begin
         if (~|chan_sel)
	    begin
	       op_sel <= op_sel << 1;
	       train_done <= 1'b0;
            chan_sel <= 16'b0000_0000_0000_0001;
	       cnt_rst <= 1'b1;
	       dataout_sel <= 16'b0000_0000_0000_0000;	    	    	     	    
	    end
	    else if (deskew_done)
	    begin
            op_sel <= op_sel;
            train_done <= 1'b0;
            chan_sel <= chan_sel << 1; 
	       cnt_rst <= 1'b1;
	       dataout_sel <= 16'b0000_0000_0000_0000;	    	    	    
	    end
	    else
	    begin
            op_sel <= op_sel;
            train_done <= 1'b0;
            chan_sel <= chan_sel; 
	       cnt_rst <= 1'b1;
	       dataout_sel <= 16'b0000_0000_0000_0000;	    	    	    
	    end
      end      
      else if (op_sel[1])
      begin
         if (~|chan_sel)
	    begin
	       op_sel <= op_sel << 1;
	       train_done <= 1'b0;
            chan_sel <= 16'b1000_0000_0000_0000;
	       cnt_rst <= 1'b0;
	       dataout_sel <= 16'b0000_0000_0000_0000;	    	    	     	    
	    end
	    else if (bitslip_done)
	    begin
            op_sel <= op_sel;
            train_done <= 1'b0;
            chan_sel <= chan_sel << 1; 
	       cnt_rst <= 1'b1;
	       dataout_sel <= 16'b0000_0000_0000_0000;	    	    	    
	     end
	    else
	    begin
            op_sel <= op_sel;
            train_done <= 1'b0;
            chan_sel <= chan_sel; 
	       cnt_rst <= 1'b1;
	       dataout_sel <= 16'b0000_0000_0000_0000;	    	    	    
	    end
      end
      else if (op_sel[2])
      begin
         if (~|chan_sel)
	    begin
	       op_sel <= op_sel << 1;
	       train_done <= 1'b1;
            chan_sel <= 16'b0000_0000_0000_0001;	    
	       cnt_rst <= 1'b1;
	       if (polarity[1])
	       begin
	          if (polarity[0])
	             dataout_sel <= dataout_sel_int;
	          else
	             dataout_sel <= ~dataout_sel_int;
	       end
	       else
	          dataout_sel <= 16'b0000_0000_0000_0000;	       		  	   	     	    	    
	    end
	    // If counter = 3, increment chan sel
	    else if (counter[1] & counter[0]) 
	    begin
            op_sel <= op_sel;
            train_done <= 1'b0;
            chan_sel <= chan_sel >> 1; 
	       cnt_rst <= 1'b0;
	       dataout_sel <= dataout_sel;	    	    	    
	    end
	    else
	    begin
            op_sel <= op_sel;
            train_done <= 1'b0;
            chan_sel <= chan_sel; 
	       cnt_rst <= 1'b0;
	       dataout_sel <= dataout_sel;	    	    	    
	    end
      end
      else if (op_sel[3])
      begin
	    op_sel <= op_sel;
	    train_done <= 1'b1;
	    cnt_rst <= 1'b0;
	    dataout_sel <= dataout_sel;	
         if (~|chan_sel)
            chan_sel <= 16'b0000_0000_0000_0001;    	    	     	    
	    else if (monitor_done)
            chan_sel <= chan_sel << 1; 	    	    	    
	    else
	    begin
            chan_sel <= chan_sel; 	    	    	    
	    end
	 end
      else
      begin
         op_sel <= 4'b0000;
	    train_done <= 1'b0;
         chan_sel <= 16'b0000_0000_0000_0000;
	    cnt_rst <= 1'b1;
	    dataout_sel <= 16'b0000_0000_0000_0000;
	 end
   end
   else // !train_en
   begin 
      if (op_sel[3])
      begin
	    op_sel <= op_sel;
	    train_done <= 1'b1;
	    cnt_rst <= 1'b0;
	    dataout_sel <= dataout_sel;	
         if (~|chan_sel)
            chan_sel <= 16'b0000_0000_0000_0001;    	    	     	    
	    else if (monitor_done)
            chan_sel <= chan_sel << 1; 	    	    	    
	    else
	    begin
            chan_sel <= chan_sel; 	    	    	    
	    end
      end
      else
      begin
         op_sel <= 4'b0000;
	    train_done <= 1'b0;
         chan_sel <= 16'b0000_0000_0000_0000;
	    cnt_rst <= 1'b1;
	    dataout_sel <= 16'b0000_0000_0000_0000;
      end
   end
end		

// Dataout Channel Select Mux
assign muxout_d0d1 = chan_sel[1] ? dataout1_R1 : dataout0_R1;
assign muxout_d2d3 = chan_sel[3] ? dataout3_R1 : dataout2_R1;
assign muxout_d4d5 = chan_sel[5] ? dataout5_R1 : dataout4_R1;
assign muxout_d6d7 = chan_sel[7] ? dataout7_R1 : dataout6_R1;
assign muxout_d8d9 = chan_sel[9] ? dataout9_R1 : dataout8_R1;
assign muxout_d10d11 = chan_sel[11] ? dataout11_R1 : dataout10_R1;
assign muxout_d12d13 = chan_sel[13] ? dataout13_R1 : dataout12_R1;
assign muxout_d14d15 = chan_sel[15] ? dataout15_R1 : dataout14_R1;

assign muxout_d0_to_d3 = (chan_sel[2] | chan_sel[3]) ?  muxout_d2d3: muxout_d0d1;
assign muxout_d4_to_d7 = (chan_sel[6] | chan_sel[7]) ?  muxout_d6d7: muxout_d4d5;
assign muxout_d8_to_d11 = (chan_sel[10] | chan_sel[11]) ?  muxout_d10d11: muxout_d8d9;
assign muxout_d12_to_d15 = (chan_sel[14] | chan_sel[15]) ?  muxout_d14d15: muxout_d12d13;

// Pipeline Registers inserted in this stage
assign muxout_d0_to_d7 = (chan_sel[4] | chan_sel[5] | chan_sel[6] | chan_sel[7]) ?  muxout_d4_to_d7_R1: muxout_d0_to_d3_R1;
assign muxout_d8_to_d15 = (chan_sel[12] | chan_sel[13] | chan_sel[14] | chan_sel[15]) ?  muxout_d12_to_d15_R1: muxout_d8_to_d11_R1;

assign datain = (chan_sel[8] | chan_sel[9] | chan_sel[10] | chan_sel[11] | 
       chan_sel[12] | chan_sel[13] | chan_sel[14] | chan_sel[15]) ?  
       muxout_d8_to_d15: muxout_d0_to_d7;

// Window Monitor Channel Select Mux
assign muxout_w0w1 = chan_sel[1] ? windowout1_R1 : windowout0_R1;
assign muxout_w2w3 = chan_sel[3] ? windowout3_R1 : windowout2_R1;
assign muxout_w4w5 = chan_sel[5] ? windowout5_R1 : windowout4_R1;
assign muxout_w6w7 = chan_sel[7] ? windowout7_R1 : windowout6_R1;
assign muxout_w8w9 = chan_sel[9] ? windowout9_R1 : windowout8_R1;
assign muxout_w10w11 = chan_sel[11] ? windowout11_R1 : windowout10_R1;
assign muxout_w12w13 = chan_sel[13] ? windowout13_R1 : windowout12_R1;
assign muxout_w14w15 = chan_sel[15] ? windowout15_R1 : windowout14_R1;

assign muxout_w0_to_w3 = (chan_sel[2] | chan_sel[3]) ?  muxout_w2w3: muxout_w0w1;
assign muxout_w4_to_w7 = (chan_sel[6] | chan_sel[7]) ?  muxout_w6w7: muxout_w4w5;
assign muxout_w8_to_w11 = (chan_sel[10] | chan_sel[11]) ?  muxout_w10w11: muxout_w8w9;
assign muxout_w12_to_w15 = (chan_sel[14] | chan_sel[15]) ?  muxout_w14w15: muxout_w12w13;

// Pipeline Registers inserted in this stage
assign muxout_w0_to_w7 = (chan_sel[4] | chan_sel[5] | chan_sel[6] | chan_sel[7]) ?  muxout_w4_to_w7_R1: muxout_w0_to_w3_R1;
assign muxout_w8_to_w15 = (chan_sel[12] | chan_sel[13] | chan_sel[14] | chan_sel[15]) ?  muxout_w12_to_w15_R1: muxout_w8_to_w11_R1;

assign datain_monitor = (chan_sel[8] | chan_sel[9] | chan_sel[10] | chan_sel[11] | 
       chan_sel[12] | chan_sel[13] | chan_sel[14] | chan_sel[15]) ?  
       muxout_w8_to_w15: muxout_w0_to_w7;


// Pipline Registers for Mux
always @(posedge rst or posedge clkdiv)
begin
   if (rst)
   begin	
      muxout_d0_to_d3_R1 <= 4'b0;   
      muxout_d4_to_d7_R1 <= 4'b0;
      muxout_d8_to_d11_R1 <= 4'b0;
      muxout_d12_to_d15_R1 <= 4'b0;

      muxout_w0_to_w3_R1 <= 4'b0;   
      muxout_w4_to_w7_R1 <= 4'b0;
      muxout_w8_to_w11_R1 <= 4'b0;
      muxout_w12_to_w15_R1 <= 4'b0;

   end
   else
   begin
      muxout_d0_to_d3_R1 <= muxout_d0_to_d3;   
      muxout_d4_to_d7_R1 <= muxout_d4_to_d7;
      muxout_d8_to_d11_R1 <= muxout_d8_to_d11;
      muxout_d12_to_d15_R1 <= muxout_d12_to_d15;

      muxout_w0_to_w3_R1 <= muxout_w0_to_w3;   
      muxout_w4_to_w7_R1 <= muxout_w4_to_w7;
      muxout_w8_to_w11_R1 <= muxout_w8_to_w11;
      muxout_w12_to_w15_R1 <= muxout_w12_to_w15;

   end
end

// Instantiate muxes that route the bitslip, dlyce and dlyinc control signals to 
// the data ISERDES and monitor ISERDES.
generate
genvar k;
   for (k=0; k<=num_chan-1; k=k+1)
   begin :  bitslip_generate
      assign bitslip[k] = chan_sel[k] ? bitslip_training : 1'b0;	
      assign dlyce[k] = chan_sel[k] ? dlyce1 : 1'b0;
      assign dlyinc[k] = chan_sel[k] ? dlyinc1 : 1'b0;	
      assign dlyce_iserdes_window[k] = chan_sel[k] ? dlyce_iserdes_window1 : 1'b0;
      assign dlyinc_iserdes_window[k] = chan_sel[k] ? dlyinc_iserdes_window1 : 1'b0; 	 	 		   

      // If counter = 2 and no edge found, select chan_sel
      always @(posedge rst or posedge clkdiv)
         if (rst)
            chan_sel1[k] <= 1'b0;
         else if ((counter[1] && ~counter[0]) && ~edgei_training[1])
            chan_sel1[k] <= chan_sel[k];	        
         else
            chan_sel1[k] <= 1'b0;	  	          

      always @(posedge rst or posedge clkdiv)
         if (rst)
	         dataout_sel_int[k] <= 1'b0;  
	      else	       
	         dataout_sel_int[k] <= dataout_sel_int[k] | chan_sel1[k];		   	    	 
   end 
endgenerate

always @(posedge rst or posedge clkdiv)
   if (rst)
      polarity <= 2'b0;	
   else if ((counter[1] && ~counter[0]) && ~edgei_training[1])
   begin
      polarity[1] <= 1'b1;
      polarity[0] <= datain[3];	
   end	        

assign dlyce1 = train_done ? dlyce_monitor : dlyce_training;
assign dlyce_iserdes_window1 = train_done ? dlyce_window : dlyce_training;			   
assign dlyinc1 = train_done ? dlyinc_monitor : dlyinc_training;
assign dlyinc_iserdes_window1 = train_done ? dlyinc_window : dlyinc_training;	
 	  	    			   
endmodule
