--******************************************************************************
--
--  Xilinx, Inc. 2002                 www.xilinx.com
--
--
--*******************************************************************************
--
--  File name :       testb_16ch_training_monitor.vhd
--
--  Description :     This module implements the testbench for the DPA design
-- 
--                    
--  Date - revision : 11/30/2004
--
--  Author :          Tze Yi Yeoh
--
--  Contact : e-mail  hotline@xilinx.com
--            phone   + 1 800 255 7778 
--
--  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are 
--              provided to you "as is". Xilinx and its licensors make and you 
--              receive no warranties or conditions, express, implied, 
--              statutory or otherwise, and Xilinx specifically disclaims any 
--              implied warranties of merchantability, non-infringement, or 
--              fitness for a particular purpose. Xilinx does not warrant that 
--              the functions contained in these designs will meet your 
--              requirements, or that the operation of these designs will be 
--              uninterrupted or error free, or that defects in the Designs 
--              will be corrected. Furthermore, Xilinx does not warrant or 
--              make any representations regarding use or the results of the 
--              use of the designs in terms of correctness, accuracy, 
--              reliability, or otherwise. 
--
--              LIMITATION OF LIABILITY. In no event will Xilinx or its 
--              licensors be liable for any loss of data, lost profits, cost 
--              or procurement of substitute goods or services, or for any 
--              special, incidental, consequential, or indirect damages 
--              arising from the use or operation of the designs or 
--              accompanying documentation, however caused and on any theory 
--              of liability. This limitation will apply even if Xilinx 
--              has been advised of the possibility of such damage. This 
--              limitation shall apply not-withstanding the failure of the 
--              essential purpose of any limited remedies herein. 
--
--  Copyright  2002 Xilinx, Inc.
--  All rights reserved 
-- 
--*****************************************************************************


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.Std_logic_arith.all;
use IEEE.Std_logic_unsigned.all;
use IEEE.numeric_std.all;
use IEEE.numeric_bit.all;
use work.all;

library unisim;
use unisim.vcomponents.all;

ENTITY testb_16ch_training_monitor IS
END testb_16ch_training_monitor;

ARCHITECTURE testb_arch OF testb_16ch_training_monitor IS

   COMPONENT top_16ch_training_monitor
      PORT (
         dataout                 : OUT std_logic_vector(3 DOWNTO 0);
         train_done              : OUT std_logic;
         ready                   : OUT std_logic;
         clkin                   : IN  std_logic;
         clkdlyctrl              : IN  std_logic;
         dip                     : IN  std_logic_vector(15 DOWNTO 0);
         din                     : IN  std_logic_vector(15 DOWNTO 0);
         dlyrst                  : IN  std_logic;
         dlyctrlrst              : IN  std_logic;
         rst                     : IN  std_logic;
         train_en                : IN  std_logic);
   END COMPONENT;


   SIGNAL dataout                  :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL done                     :  std_logic;   
   SIGNAL ready                    :  std_logic;   
   SIGNAL train_done               :  std_logic;   
   SIGNAL clkin                    :  std_logic;   
   SIGNAL clkdlyctrl               :  std_logic;   
   SIGNAL dip                      :  std_logic_vector(15 DOWNTO 0);   
   SIGNAL din                      :  std_logic_vector(15 DOWNTO 0);   
   SIGNAL txdata                   :  std_logic;   
   SIGNAL dlyrst                   :  std_logic; 
   SIGNAL dlyctrlrst               :  std_logic;       
   SIGNAL rst                      :  std_logic;   
   SIGNAL train_en                 :  std_logic;   
   SIGNAL test                     :  std_logic_vector(19 DOWNTO 0);   
   SIGNAL realdata                 :  std_logic_vector(19 DOWNTO 0); 
   SIGNAL realdata_start           :  std_logic;      
   SIGNAL counter                  :  integer;   
   SIGNAL R                        :  real;   
   CONSTANT  TEST_VECTOR           :  std_logic_vector(19 DOWNTO 0) := "00000000001111111111";    
   CONSTANT  REAL_DATA             :  std_logic_vector(19 DOWNTO 0) := X"55555";    
   CONSTANT  tjmax                 :  integer := 300;   
   CONSTANT tHCLK                  :  TIME := 1.0 ns; -- Input clock frequency
   CONSTANT tREFCLK                :  TIME := 2.50 ns; -- Input clock frequency
    

BEGIN
   
   -- Instantiate top level module
   dut : top_16ch_training_monitor 
      PORT MAP (
         dataout => dataout,
         train_done => done,
         ready => ready,
         clkin => clkin,
         clkdlyctrl => clkdlyctrl,
         dip => dip,
         din => din,
         dlyrst => dlyrst,
         dlyctrlrst => dlyctrlrst,
         rst => rst,
         train_en => train_en);   
   
   din <= NOT dip ;

   -- Input forwarded source synchronous clock      
   PROCESS
   BEGIN      
      clkin <= '0';
	 WAIT FOR 1.0 ns;	 	 	    
      WHILE (TRUE) LOOP
         clkin <= NOT clkin;
	    WAIT FOR tHCLK;    
      END LOOP;
   END PROCESS;

   -- 200 Mhz REFCLK for IDELAYCTRL module
   PROCESS
   BEGIN
      clkdlyctrl <= '0';
	 WAIT FOR 1.0 ns;	     
      WHILE (TRUE) LOOP
         clkdlyctrl <= NOT clkdlyctrl;
	    WAIT FOR tREFCLK;    
      END LOOP;
   END PROCESS;

   PROCESS
   BEGIN        
	 train_en <= '0';             
	 rst <= '0';
	 dlyrst <= '0'; 
	 dlyctrlrst <= '0'; 
	 WAIT FOR 10 ns;
	 rst <= '1';
	 dlyrst <= '1'; 
	 dlyctrlrst <= '1';
	 WAIT FOR 10 ns;
	 rst <= '0';
	 dlyrst <= '0'; 
	 dlyctrlrst <= '0';	 	
	 WAIT FOR 70 ns;
	 train_en <= '1';	  	 	      
  	 WAIT UNTIL done = '1';
	 train_en <= '0';
	 WAIT;
   END PROCESS;

   -- Generate input serial data stream   
   PROCESS (clkin, rst, realdata_start)
   BEGIN
      IF (rst = '1') THEN
         txdata <= '0';   
         test <= TEST_VECTOR;    
         realdata <= REAL_DATA; 	     
      ELSIF (rising_edge(clkin) OR falling_edge(clkin)) THEN
         -- At reset, send training pattern         
         IF (train_done = '0') THEN           
            txdata <= test(19);    
            test <= test(18 DOWNTO 0) & test(19);    
         ELSIF (realdata_start = '1') THEN		  
            -- After training is done, send real data (oscillating signal)             
            txdata <= realdata(19);    
            realdata <= realdata(18 DOWNTO 0) & realdata(19);    
         END IF;
      END IF;
   END PROCESS;

   -- Generate input serial data stream   
   PROCESS (clkin, rst, train_done)
   BEGIN
      IF (rst = '1') THEN
         realdata_start <= '0';    	     
      ELSIF (rising_edge(clkin) OR falling_edge(clkin)) THEN
         -- At reset, send training pattern         
         IF (train_done = '0') THEN           
            realdata_start <= '0';   
         ELSIF (train_done = '1' AND counter = 3) THEN		  
            -- After training is done, send real data (oscillating signal)             
            realdata_start <= '1'; 
	    ELSE
            realdata_start <= realdata_start; 	    	     
         END IF;
      END IF;
   END PROCESS;

   transmit_gen : for K in 0 to 13 generate
   BEGIN   
      PROCESS (clkin, rst)
      BEGIN
         IF (rst = '1') THEN
            dip(K) <= '0';    
         ELSIF (rising_edge(clkin) OR falling_edge(clkin)) THEN
            dip(K) <= TRANSPORT txdata AFTER 19.0 ns;    
         END IF;
      END PROCESS;
   END GENERATE;

   -- Skew Channel 1 by -1 bit period   
   PROCESS (clkin, rst)
   BEGIN
      IF (rst = '1') THEN
         dip(14) <= '0';    
      ELSIF (rising_edge(clkin) OR falling_edge(clkin)) THEN
         dip(14) <= transport txdata AFTER 20.0 ns;    
      END IF;
   END PROCESS;

   -- Skew Channel 0 by +1 bit period   
   PROCESS (clkin, rst)
   BEGIN
      IF (rst = '1') THEN
         dip(15) <= '0';    
      ELSIF (rising_edge(clkin) OR falling_edge(clkin)) THEN
         dip(15) <= transport txdata AFTER 21.0 ns;    
      END IF;
   END PROCESS;

   -- R variable used to simulate jitter.   
   PROCESS (clkin, rst)
   BEGIN
      IF (rst = '1') THEN
         R <= 0.0;    
         counter <= 0;    
      ELSIF (rising_edge(clkin) OR falling_edge(clkin)) THEN
         IF (counter = 4) THEN
            counter <= 1;    
            R <= 1.0;    
         ELSE
            R <= 0.0;    
            counter <= counter + 1;    
         END IF;
      END IF;
   END PROCESS;

   PROCESS (done, rst)
   BEGIN
      IF (rst = '1') THEN
         train_done <= '0'; 
      ELSIF (done = '1') THEN
         train_done <= '1';    
      ELSE
         train_done <= '0';    
      END IF;
   END PROCESS;

END testb_arch;
