--******************************************************************************
--
--  Xilinx, Inc. 2002                 www.xilinx.com
--
--
--*******************************************************************************
--
--  File name :       top_16ch_training_monitor.vhd
--
--  Description :     This module is the top level module for the DPA design
-- 
--                    
--  Date - revision : 11/30/2004
--
--  Author :          Tze Yi Yeoh
--
--  Contact : e-mail  hotline@xilinx.com
--            phone   + 1 800 255 7778 
--
--  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are 
--              provided to you "as is". Xilinx and its licensors make and you 
--              receive no warranties or conditions, express, implied, 
--              statutory or otherwise, and Xilinx specifically disclaims any 
--              implied warranties of merchantability, non-infringement, or 
--              fitness for a particular purpose. Xilinx does not warrant that 
--              the functions contained in these designs will meet your 
--              requirements, or that the operation of these designs will be 
--              uninterrupted or error free, or that defects in the Designs 
--              will be corrected. Furthermore, Xilinx does not warrant or 
--              make any representations regarding use or the results of the 
--              use of the designs in terms of correctness, accuracy, 
--              reliability, or otherwise. 
--
--              LIMITATION OF LIABILITY. In no event will Xilinx or its 
--              licensors be liable for any loss of data, lost profits, cost 
--              or procurement of substitute goods or services, or for any 
--              special, incidental, consequential, or indirect damages 
--              arising from the use or operation of the designs or 
--              accompanying documentation, however caused and on any theory 
--              of liability. This limitation will apply even if Xilinx 
--              has been advised of the possibility of such damage. This 
--              limitation shall apply not-withstanding the failure of the 
--              essential purpose of any limited remedies herein. 
--
--  Copyright  2002 Xilinx, Inc.
--  All rights reserved 
-- 
--*****************************************************************************

library IEEE;
use IEEE.std_logic_1164.all;
--use IEEE.Std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use IEEE.numeric_std.all;
--use IEEE.numeric_bit.all;
use work.all;

library unisim;
use unisim.vcomponents.all;

ENTITY top_16ch_training_monitor IS
   PORT (
      dataout                 : OUT std_logic_vector(3 DOWNTO 0);   
      train_done              : OUT std_logic;   
      ready                   : OUT std_logic;   
      clkin                   : IN std_logic;   
      clkdlyctrl              : IN std_logic;   
      dip                     : IN std_logic_vector(15 DOWNTO 0);   
      din                     : IN std_logic_vector(15 DOWNTO 0);   
      dlyrst                  : IN std_logic;   
      dlyctrlrst              : IN std_logic;   
      rst                     : IN std_logic;   
      train_en                : IN std_logic);   
END ENTITY top_16ch_training_monitor;

ARCHITECTURE top_arch OF top_16ch_training_monitor IS

   component IDELAYCTRL
   port(
      RDY: out std_logic;
      REFCLK: in std_logic;
      RST: in std_logic
   );
   end component;

   component IBUFG
   port(
      O: out std_logic;
      I: in std_logic
   );
   end component;  
   
   component BUFG
   port(
      O: out std_logic;
      I: in std_logic
   );
   end component;  
   
   component IBUF
   port(
      O: out std_logic;
      I: in std_logic
   );
   end component; 

   component BUFIO
   port(
      O: out std_logic;
      I: in std_logic
   );
   end component;    
      
   component BUFR
   -- synthesis translate_off
   generic(
      BUFR_DIVIDE : string := "BYPASS"
   );
   -- synthesis translate_on
   port(
      O: out std_ulogic;
      CE: in std_ulogic;
      CLR: in std_ulogic;
      I: in std_ulogic
   );
   end component;

   component IBUFDS_DIFF_OUT
   port(
      I: in std_logic;
      IB: in std_logic;
      O: out std_logic;
      OB: out std_logic
   );
   end component; 

   component ISERDES
   -- synthesis translate_off
   generic (
      BITSLIP_ENABLE : boolean := FALSE; --(TRUE, FALSE)
      DATA_RATE : string := "DDR"; --(SDR, DDR)
      DATA_WIDTH : integer := 4; --(2,3,4,5,6,7,8,10)
      INTERFACE_TYPE : string := "MEMORY"; --(MEMORY, NETWORKING)
      IOBDELAY : string := "NONE"; --(NONE,IBUF,IFD,BOTH)
      IOBDELAY_TYPE : string := "DEFAULT"; --(DEFAULT,FIXED, VARIABLE)
      IOBDELAY_VALUE : integer := 0; --(0 to 63)
      NUM_CE : integer := 2; --(1,2)
      SERDES_MODE : string := "MASTER" --(MASTER, SLAVE)
   );
   -- synthesis translate_on
   port (
      O : out STD_LOGIC;
      Q1 : out STD_LOGIC;
      Q2 : out STD_LOGIC;
      Q3 : out STD_LOGIC;
      Q4 : out STD_LOGIC;
      Q5 : out STD_LOGIC;
      Q6 : out STD_LOGIC;
      SHIFTOUT1 : out STD_LOGIC;
      SHIFTOUT2 : out STD_LOGIC;
      BITSLIP : in STD_LOGIC;
      CE1 : in STD_LOGIC;
      CE2 : in STD_LOGIC;
      CLK : in STD_LOGIC;
      CLKDIV : in STD_LOGIC;
      D : in STD_LOGIC;
      DLYCE : in STD_LOGIC;
      DLYINC : in STD_LOGIC;
      DLYRST : in STD_LOGIC;
      OCLK : in STD_LOGIC;
      REV : in STD_LOGIC;
      SHIFTIN1 : in STD_LOGIC;
      SHIFTIN2 : in STD_LOGIC;
      SR : in STD_LOGIC
      );
   end component;

   -- Instantiate the control module that implements the word alignment algorithm
   COMPONENT bitslip_ctrl
      PORT (
         datain                  : IN  std_logic_vector(3 DOWNTO 0);
         rst                     : IN  std_logic;
         bitslip_en              : IN  std_logic;
         clkdiv                  : IN  std_logic;
         bitslip                 : OUT std_logic;
         done                    : OUT std_logic;
         error                   : OUT std_logic);
   END COMPONENT;

   -- Instantiate deskew module that implements bit alignment algorithm
   COMPONENT deskew
      PORT (
         datain                  : IN  std_logic_vector(3 DOWNTO 0);
         edgei                   : OUT std_logic_vector(2 DOWNTO 0);
         rst                     : IN  std_logic;
         deskew_en               : IN  std_logic;
         clkdiv                  : IN  std_logic;
         dlyce                   : OUT std_logic;
         dlyinc                  : OUT std_logic;
         done                    : OUT std_logic);
   END COMPONENT;

   -- Instantiate window monitoring module that implements the real-time monitoring algorithm
   COMPONENT monitor_ctrl
      PORT (
         datain                  : IN  std_logic_vector(3 DOWNTO 0);
         datain_monitor          : IN  std_logic_vector(3 DOWNTO 0);
         rst                     : IN  std_logic;
         monitor_en              : IN  std_logic;
         clkdiv                  : IN  std_logic;
         dlyce_window            : OUT std_logic;
         dlyinc_window           : OUT std_logic;
         dlyce                   : OUT std_logic;
         dlyinc                  : OUT std_logic;
         monitor_done            : OUT std_logic);
   END COMPONENT;

   --Declaring BUFR attributes  
   attribute BUFR_DIVIDE : string;
   attribute BUFR_TYPE : string;
   -- BUFR attributes
   attribute BUFR_DIVIDE of bufr_module:label is "2";
   attribute BUFR_TYPE of bufr_module:label is "NONE";

   SIGNAL data_in                  :  std_logic_vector(dip'RANGE);   
   SIGNAL window_in                :  std_logic_vector(dip'RANGE);   
   SIGNAL chan_sel                 :  std_logic_vector(dip'RANGE);   
   SIGNAL chan_sel1                :  std_logic_vector(dip'RANGE);   
   SIGNAL dataout_sel              :  std_logic_vector(dip'RANGE);   
   SIGNAL dataout_sel_int          :  std_logic_vector(dip'RANGE);   
   SIGNAL polarity                 :  std_logic_vector(1 DOWNTO 0);   
   SIGNAL deskew_done              :  std_logic;   
   SIGNAL bitslip_done             :  std_logic;   
   SIGNAL monitor_done             :  std_logic;   
   SIGNAL bitslip_error            :  std_logic;   
   SIGNAL dlyce                    :  std_logic_vector(dip'RANGE);   
   SIGNAL dlyinc                   :  std_logic_vector(dip'RANGE);   
   SIGNAL dlyce_iserdes_window     :  std_logic_vector(dip'RANGE);   
   SIGNAL dlyinc_iserdes_window    :  std_logic_vector(dip'RANGE);   
   SIGNAL dlyce_training           :  std_logic;   
   SIGNAL dlyinc_training          :  std_logic;   
   SIGNAL dlyce_window             :  std_logic;   
   SIGNAL dlyinc_window            :  std_logic;   
   SIGNAL dlyce_iserdes_window1    :  std_logic;   
   SIGNAL dlyinc_iserdes_window1   :  std_logic;   
   SIGNAL dlyce1                   :  std_logic;   
   SIGNAL dlyinc1                  :  std_logic;   
   SIGNAL bitslip                  :  std_logic_vector(dip'RANGE);   
   SIGNAL bitslip_training         :  std_logic;   
   SIGNAL iobclk                   :  std_logic;   
   SIGNAL iobclk_ibuf              :  std_logic;   
   SIGNAL clkdiv                   :  std_logic;   
   SIGNAL refclk                   :  std_logic;   
   SIGNAL refclk_ibufg             :  std_logic;   
   SIGNAL q1                       :  std_logic_vector(dip'RANGE);   
   SIGNAL q2                       :  std_logic_vector(dip'RANGE);   
   SIGNAL q3                       :  std_logic_vector(dip'RANGE);   
   SIGNAL q4                       :  std_logic_vector(dip'RANGE);   
   SIGNAL w1                       :  std_logic_vector(dip'RANGE);   
   SIGNAL w2                       :  std_logic_vector(dip'RANGE);   
   SIGNAL w3                       :  std_logic_vector(dip'RANGE);   
   SIGNAL w4                       :  std_logic_vector(dip'RANGE);   
   SIGNAL datain                   :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL datain_monitor           :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout15                :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout14                :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout13                :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout12                :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout11                :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout10                :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout9                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout8                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout7                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout6                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout5                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout4                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout3                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout2                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout1                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout0                 :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout15_R1             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout14_R1             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout13_R1             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout12_R1             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout11_R1             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout10_R1             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout9_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout8_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout7_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout6_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout5_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout4_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout3_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout2_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout1_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout0_R1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int15            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int14            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int13            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int12            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int11            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int10            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int9             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int8             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int7             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int6             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int5             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int4             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int3             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int2             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int1             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL dataout_int0             :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout15              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout14              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout13              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout12              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout11              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout10              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout9               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout8               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout7               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout6               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout5               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout4               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout3               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout2               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout1               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout0               :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout15_R1           :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout14_R1           :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout13_R1           :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout12_R1           :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout11_R1           :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout10_R1           :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout9_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout8_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout7_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout6_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout5_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout4_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout3_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout2_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout1_R1            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL windowout0_R1            :  std_logic_vector(3 DOWNTO 0);   
   -- Dataout Channel Select Mux
   SIGNAL muxout_d0d1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d2d3              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d4d5              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d6d7              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d8d9              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d10d11            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d12d13            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d14d15            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d0_to_d3          :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d4_to_d7          :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d8_to_d11         :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d12_to_d15        :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d0_to_d3_R1       :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d4_to_d7_R1       :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d8_to_d11_R1      :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d12_to_d15_R1     :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d0_to_d7          :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_d8_to_d15         :  std_logic_vector(3 DOWNTO 0);   
   -- Window Monitor Channel Select Mux
   SIGNAL muxout_w0w1              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w2w3              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w4w5              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w6w7              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w8w9              :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w10w11            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w12w13            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w14w15            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w0_to_w3          :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w4_to_w7          :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w8_to_w11         :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w12_to_w15        :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w0_to_w3_R1       :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w4_to_w7_R1       :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w8_to_w11_R1      :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w12_to_w15_R1     :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w0_to_w7          :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL muxout_w8_to_w15         :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL edgei_training           :  std_logic_vector(2 DOWNTO 0);   
   SIGNAL counter                  :  std_logic_vector(7 DOWNTO 0);   
   SIGNAL cnt_rst                  :  std_logic;   
   SIGNAL op_sel                   :  std_logic_vector(3 DOWNTO 0);  
   SIGNAL op_sel_red_or            :  std_logic; 
   SIGNAL chan_sel_red_or          :  std_logic; 
   -- Declare parameter for channel select 
   CONSTANT  CHAN0                 :  std_logic_vector(dip'RANGE) := "0000000000000001";    
   CONSTANT  CHAN1                 :  std_logic_vector(dip'RANGE) := "0000000000000010";    
   CONSTANT  CHAN2                 :  std_logic_vector(dip'RANGE) := "0000000000000100";    
   CONSTANT  CHAN3                 :  std_logic_vector(dip'RANGE) := "0000000000001000";    
   CONSTANT  CHAN4                 :  std_logic_vector(dip'RANGE) := "0000000000010000";    
   CONSTANT  CHAN5                 :  std_logic_vector(dip'RANGE) := "0000000000100000";    
   CONSTANT  CHAN6                 :  std_logic_vector(dip'RANGE) := "0000000001000000";    
   CONSTANT  CHAN7                 :  std_logic_vector(dip'RANGE) := "0000000010000000";    
   CONSTANT  CHAN8                 :  std_logic_vector(dip'RANGE) := "0000000100000000";    
   CONSTANT  CHAN9                 :  std_logic_vector(dip'RANGE) := "0000001000000000";    
   CONSTANT  CHAN10                :  std_logic_vector(dip'RANGE) := "0000010000000000";    
   CONSTANT  CHAN11                :  std_logic_vector(dip'RANGE) := "0000100000000000";    
   CONSTANT  CHAN12                :  std_logic_vector(dip'RANGE) := "0001000000000000";    
   CONSTANT  CHAN13                :  std_logic_vector(dip'RANGE) := "0010000000000000";    
   CONSTANT  CHAN14                :  std_logic_vector(dip'RANGE) := "0100000000000000";    
   CONSTANT  CHAN15                :  std_logic_vector(dip'RANGE) := "1000000000000000";    

   SIGNAL deskew_en              :  std_logic;   

   SIGNAL bitslip_en              :  std_logic;   
   SIGNAL dataout_int            :  std_logic_vector(3 DOWNTO 0);   
   SIGNAL train_done_int         :  std_logic;   
   SIGNAL ready_int              :  std_logic;   
   SIGNAL dlyce_monitor            :  std_logic;   
   SIGNAL dlyinc_monitor           :  std_logic;   

BEGIN
   dataout <= dataout_int;
   train_done <= train_done_int;
   ready <= ready_int;
   op_sel_red_or <= op_sel(3) OR op_sel(2) OR op_sel(1) OR op_sel(0);
   chan_sel_red_or <= chan_sel(15) OR chan_sel(14) OR chan_sel(13) OR chan_sel(12) OR
                      chan_sel(11) OR chan_sel(10) OR chan_sel(9) OR chan_sel(8) OR
				  chan_sel(7) OR chan_sel(6) OR chan_sel(5) OR chan_sel(4) OR
				  chan_sel(3) OR chan_sel(2) OR chan_sel(1) OR chan_sel(0);     
   -- Instantiate IDELAYCTRL module because the ISERDES is used in    
   -- VARIABLE tap delay chain mode. 
   dlyctrl : IDELAYCTRL 
      PORT MAP (
         RDY => ready_int,
         REFCLK => refclk,
         RST => dlyctrlrst);   
   
   ibufg_module : IBUFG 
      PORT MAP (
         O => refclk_ibufg,
         I => clkdlyctrl);   
   
   bufg_module : BUFG 
      PORT MAP (
         O => refclk,
         I => refclk_ibufg);   
   
   ibuf_module : IBUF 
      PORT MAP (
         O => iobclk_ibuf,
         I => clkin);   
      
   -- Instantiate regional clock buffers. One regional clock buffer   
   -- can accomodate up to 3 clock regions (96 IOBs). 
   bufio_module : BUFIO 
      PORT MAP (
         O => iobclk,
         I => iobclk_ibuf);   
	        
   -- Note that the clock divide is set to 2 not 4 for a 1:4    
   -- deserialization factor in DDR mode.
   bufr_module : BUFR
   -- synthesis translate_off
      GENERIC MAP (
         BUFR_DIVIDE => "2")
   -- synthesis translate_on
      Port map (
         O => clkdiv,
         CE => '1',
         CLR => '0',
         I => iobclk
	    ); 

   channels:for K in dip'length-1 downto 0 generate
   -- Component Attribute specification for ISERDES
   attribute BITSLIP_ENABLE : boolean;
   attribute DATA_RATE : string;
   attribute DATA_WIDTH : integer;
   attribute INTERFACE_TYPE : string;
   attribute IOBDELAY : string;
   attribute IOBDELAY_TYPE : string;
   attribute IOBDELAY_VALUE : integer;
   attribute NUM_CE : integer;
   attribute SERDES_MODE : string;

      attribute BITSLIP_ENABLE of data_chan_master:label is TRUE;
      attribute DATA_RATE of data_chan_master:label is "DDR";
      attribute DATA_WIDTH of data_chan_master:label is 4;
      attribute INTERFACE_TYPE of data_chan_master:label is "NETWORKING";
      attribute IOBDELAY of data_chan_master:label is "IFD";
      attribute IOBDELAY_TYPE of data_chan_master:label is "VARIABLE";
      attribute IOBDELAY_VALUE of data_chan_master:label is 0;
      attribute NUM_CE of data_chan_master:label is 1;
      attribute SERDES_MODE of data_chan_master:label is "MASTER";

      attribute BITSLIP_ENABLE of window_monitor:label is TRUE;
      attribute DATA_RATE of window_monitor:label is "DDR";
      attribute DATA_WIDTH of window_monitor:label is 4;
      attribute INTERFACE_TYPE of window_monitor:label is "NETWORKING";
      attribute IOBDELAY of window_monitor:label is "IFD";
      attribute IOBDELAY_TYPE of window_monitor:label is "VARIABLE";
      attribute IOBDELAY_VALUE of window_monitor:label is 3;
      attribute NUM_CE of window_monitor:label is 1;
      attribute SERDES_MODE of window_monitor:label is "MASTER";
   BEGIN  
      -- Instantiate data differential input buffers
      datain_buf : IBUFDS_DIFF_OUT 
         PORT MAP (
            I => dip(K),
            IB => din(K),
            O => data_in(K),
            OB => window_in(K));   
      
      -- Instantiate the master data channel ISERDES
      data_chan_master : ISERDES 
      -- synthesis translate_off
      generic map (
         BITSLIP_ENABLE => TRUE, --(TRUE, FALSE)
         DATA_RATE => "DDR", --(SDR, DDR)
         DATA_WIDTH => 4, --(2,3,4,5,6,7,8,10)
         INTERFACE_TYPE => "NETWORKING", --(MEMORY, NETWORKING)
         IOBDELAY => "IFD", --(NONE,IBUF,IFD,BOTH)
         IOBDELAY_TYPE => "VARIABLE", --(DEFAULT, FIXED, VARIABLE)
         IOBDELAY_VALUE => 6, --(0 to 63)
         NUM_CE => 1, --(1,2)
         SERDES_MODE => "MASTER" --(MASTER, SLAVE)
         )
      --synthesis translate_on
         PORT MAP (
            O => open,
            Q1 => q1(K),
            Q2 => q2(K),
            Q3 => q3(K),
            Q4 => q4(K),
            Q5 => open,
            Q6 => open,
            SHIFTOUT1 => open,
            SHIFTOUT2 => open,
            BITSLIP => bitslip(K),
            CE1 => '1',
            CE2 => '1',
            CLK => iobclk,
            CLKDIV => clkdiv,
            D => data_in(K),
            DLYCE => dlyce(K),
            DLYINC => dlyinc(K),
            DLYRST => dlyrst,
            OCLK => '0',
            REV => '0',
            SHIFTIN1 => '0',
            SHIFTIN2 => '0',
            SR => rst);   
      
      -- Instantiate the window monitor data channel ISERDES 
      window_monitor : ISERDES 
      -- synthesis translate_off
      generic map (
         BITSLIP_ENABLE => TRUE, --(TRUE, FALSE)
         DATA_RATE => "DDR", --(SDR, DDR)
         DATA_WIDTH => 4, --(2,3,4,5,6,7,8,10)
         INTERFACE_TYPE => "NETWORKING", --(MEMORY, NETWORKING)
         IOBDELAY => "IFD", --(NONE,IBUF,IFD,BOTH)
         IOBDELAY_TYPE => "VARIABLE", --(DEFAULT, FIXED, VARIABLE)
         IOBDELAY_VALUE => 9, --(0 to 63)
         NUM_CE => 1, --(1,2)
         SERDES_MODE => "MASTER" --(MASTER, SLAVE)
         )
      --synthesis translate_on
         PORT MAP (
            O => open,
            Q1 => w1(K),
            Q2 => w2(K),
            Q3 => w3(K),
            Q4 => w4(K),
            Q5 => open,
            Q6 => open,
            SHIFTOUT1 => open,
            SHIFTOUT2 => open,
            BITSLIP => bitslip(K),
            CE1 => '1',
            CE2 => '1',
            CLK => iobclk,
            CLKDIV => clkdiv,
            D => window_in(K),
            DLYCE => dlyce_iserdes_window(K),
            DLYINC => dlyinc_iserdes_window(K),
            DLYRST => dlyrst,
            OCLK => '0',
            REV => '0',
            SHIFTIN1 => '0',
            SHIFTIN2 => '0',
            SR => rst);   

      bitslip(K) <= bitslip_training WHEN chan_sel(K) = '1' ELSE '0';
	 dlyce(K) <= dlyce1 WHEN chan_sel(K) = '1' ELSE '0' ;
	 dlyinc(K) <= dlyinc1 WHEN chan_sel(K) = '1' ELSE '0' ;
	 dlyce_iserdes_window(K) <= dlyce_iserdes_window1 WHEN chan_sel(K) = '1' ELSE '0' ;
	 dlyinc_iserdes_window(K) <= dlyinc_iserdes_window1 WHEN chan_sel(K) = '1' ELSE '0' ;

      -- If counter = 2 and no edge found, select chan_sel   
      PROCESS (rst, clkdiv)
      BEGIN
         IF (rst = '1') THEN
            chan_sel1(K) <= '0';    
         ELSIF rising_edge(clkdiv) THEN
		  -- If counter = 2 and no edge found, select chan_sel
            IF (((counter(1) AND NOT counter(0)) AND NOT edgei_training(1)) = '1') THEN
               chan_sel1(K) <= chan_sel(K);    
            ELSE
               chan_sel1(K) <= '0';    
            END IF;
         END IF;
      END PROCESS;

      PROCESS (rst, clkdiv)
      BEGIN
         IF (rst = '1') THEN
            dataout_sel_int(K) <= '0';    
         ELSIF rising_edge(clkdiv)	THEN
            dataout_sel_int(K) <= dataout_sel_int(K) OR chan_sel1(K);    
         END IF;
      END PROCESS;
   END GENERATE;

   -- If counter = 2 and no edge found, select chan_sel   
   PROCESS (rst, clkdiv)
   BEGIN
      IF (rst = '1') THEN
          polarity <= "00";    
      ELSIF rising_edge(clkdiv) THEN
	      -- If counter = 2 and no edge found, select chan_sel
         IF (((counter(1) AND NOT counter(0)) AND NOT edgei_training(1)) = '1') THEN
            polarity(1) <= '1';    
            polarity(0) <= datain(3);     
         END IF;
      END IF;
   END PROCESS;
		        
   dataout15 <= q4(15) & q3(15) & q2(15) & q1(15) ;
   dataout14 <= q4(14) & q3(14) & q2(14) & q1(14) ;
   dataout13 <= q4(13) & q3(13) & q2(13) & q1(13) ;
   dataout12 <= q4(12) & q3(12) & q2(12) & q1(12) ;
   dataout11 <= q4(11) & q3(11) & q2(11) & q1(11) ;
   dataout10 <= q4(10) & q3(10) & q2(10) & q1(10) ;
   dataout9 <= q4(9) & q3(9) & q2(9) & q1(9) ;
   dataout8 <= q4(8) & q3(8) & q2(8) & q1(8) ;
   dataout7 <= q4(7) & q3(7) & q2(7) & q1(7) ;
   dataout6 <= q4(6) & q3(6) & q2(6) & q1(6) ;
   dataout5 <= q4(5) & q3(5) & q2(5) & q1(5) ;
   dataout4 <= q4(4) & q3(4) & q2(4) & q1(4) ;
   dataout3 <= q4(3) & q3(3) & q2(3) & q1(3) ;
   dataout2 <= q4(2) & q3(2) & q2(2) & q1(2) ;
   dataout1 <= q4(1) & q3(1) & q2(1) & q1(1) ;
   dataout0 <= q4(0) & q3(0) & q2(0) & q1(0) ;

   windowout15 <= w4(15) & w3(15) & w2(15) & w1(15) ;
   windowout14 <= w4(14) & w3(14) & w2(14) & w1(14) ;
   windowout13 <= w4(13) & w3(13) & w2(13) & w1(13) ;
   windowout12 <= w4(12) & w3(12) & w2(12) & w1(12) ;
   windowout11 <= w4(11) & w3(11) & w2(11) & w1(11) ;
   windowout10 <= w4(10) & w3(10) & w2(10) & w1(10) ;
   windowout9 <= w4(9) & w3(9) & w2(9) & w1(9) ;
   windowout8 <= w4(8) & w3(8) & w2(8) & w1(8) ;
   windowout7 <= w4(7) & w3(7) & w2(7) & w1(7) ;
   windowout6 <= w4(6) & w3(6) & w2(6) & w1(6) ;
   windowout5 <= w4(5) & w3(5) & w2(5) & w1(5) ;
   windowout4 <= w4(4) & w3(4) & w2(4) & w1(4) ;
   windowout3 <= w4(3) & w3(3) & w2(3) & w1(3) ;
   windowout2 <= w4(2) & w3(2) & w2(2) & w1(2) ;
   windowout1 <= w4(1) & w3(1) & w2(1) & w1(1) ;
   windowout0 <= w4(0) & w3(0) & w2(0) & w1(0) ;

   -- Dummy OR function to OR all data outputs together. Replace with user logic													
   dataout_int <= dataout_int15 OR dataout_int14 OR dataout_int13 OR dataout_int12 OR dataout_int11 OR dataout_int10 OR dataout_int9 OR dataout_int8 OR dataout_int7 OR dataout_int6 OR dataout_int5 OR dataout_int4 OR dataout_int3 OR dataout_int2 OR dataout_int1 OR dataout_int0 ;

   -- Pipeline ISERDES outputs   
   PROCESS (clkdiv, rst)
   BEGIN
      IF (rst = '1') THEN
         dataout15_R1 <= "0000";    
         dataout14_R1 <= "0000";    
         dataout13_R1 <= "0000";    
         dataout12_R1 <= "0000";    
         dataout11_R1 <= "0000";    
         dataout10_R1 <= "0000";    
         dataout9_R1 <= "0000";    
         dataout8_R1 <= "0000";    
         dataout7_R1 <= "0000";    
         dataout6_R1 <= "0000";    
         dataout5_R1 <= "0000";    
         dataout4_R1 <= "0000";    
         dataout3_R1 <= "0000";    
         dataout2_R1 <= "0000";    
         dataout1_R1 <= "0000";    
         dataout0_R1 <= "0000";  
	      
         windowout15_R1 <= "0000";    
         windowout14_R1 <= "0000";    
         windowout13_R1 <= "0000";    
         windowout12_R1 <= "0000";    
         windowout11_R1 <= "0000";    
         windowout10_R1 <= "0000";    
         windowout9_R1 <= "0000";    
         windowout8_R1 <= "0000";    
         windowout7_R1 <= "0000";    
         windowout6_R1 <= "0000";    
         windowout5_R1 <= "0000";    
         windowout4_R1 <= "0000";    
         windowout3_R1 <= "0000";    
         windowout2_R1 <= "0000";    
         windowout1_R1 <= "0000";    
         windowout0_R1 <= "0000";    
      ELSIF rising_edge(clkdiv) THEN
         dataout15_R1 <= dataout15;    
         dataout14_R1 <= dataout14;    
         dataout13_R1 <= dataout13;    
         dataout12_R1 <= dataout12;    
         dataout11_R1 <= dataout11;    
         dataout10_R1 <= dataout10;    
         dataout9_R1 <= dataout9;    
         dataout8_R1 <= dataout8;    
         dataout7_R1 <= dataout7;    
         dataout6_R1 <= dataout6;    
         dataout5_R1 <= dataout5;    
         dataout4_R1 <= dataout4;    
         dataout3_R1 <= dataout3;    
         dataout2_R1 <= dataout2;    
         dataout1_R1 <= dataout1;    
         dataout0_R1 <= dataout0; 
	       
         windowout15_R1 <= windowout15;    
         windowout14_R1 <= windowout14;    
         windowout13_R1 <= windowout13;    
         windowout12_R1 <= windowout12;    
         windowout11_R1 <= windowout11;    
         windowout10_R1 <= windowout10;    
         windowout9_R1 <= windowout9;    
         windowout8_R1 <= windowout8;    
         windowout7_R1 <= windowout7;    
         windowout6_R1 <= windowout6;    
         windowout5_R1 <= windowout5;    
         windowout4_R1 <= windowout4;    
         windowout3_R1 <= windowout3;    
         windowout2_R1 <= windowout2;    
         windowout1_R1 <= windowout1;    
         windowout0_R1 <= windowout0;    
      END IF;
   END PROCESS;

   -- Bitslip adjust Mux that selects between ISERDES and pipeline outputs
   dataout_int15 <= dataout15_R1 WHEN dataout_sel(15) = '1' ELSE dataout15 ;
   dataout_int14 <= dataout14_R1 WHEN dataout_sel(14) = '1' ELSE dataout14 ;
   dataout_int13 <= dataout13_R1 WHEN dataout_sel(13) = '1' ELSE dataout13 ;
   dataout_int12 <= dataout12_R1 WHEN dataout_sel(12) = '1' ELSE dataout12 ;
   dataout_int11 <= dataout11_R1 WHEN dataout_sel(11) = '1' ELSE dataout11 ;
   dataout_int10 <= dataout10_R1 WHEN dataout_sel(10) = '1' ELSE dataout10 ;
   dataout_int9 <= dataout9_R1 WHEN dataout_sel(9) = '1' ELSE dataout9 ;
   dataout_int8 <= dataout8_R1 WHEN dataout_sel(8) = '1' ELSE dataout8 ;
   dataout_int7 <= dataout7_R1 WHEN dataout_sel(7) = '1' ELSE dataout7 ;
   dataout_int6 <= dataout6_R1 WHEN dataout_sel(6) = '1' ELSE dataout6 ;
   dataout_int5 <= dataout5_R1 WHEN dataout_sel(5) = '1' ELSE dataout5 ;
   dataout_int4 <= dataout4_R1 WHEN dataout_sel(4) = '1' ELSE dataout4 ;
   dataout_int3 <= dataout3_R1 WHEN dataout_sel(3) = '1' ELSE dataout3 ;
   dataout_int2 <= dataout2_R1 WHEN dataout_sel(2) = '1' ELSE dataout2 ;
   dataout_int1 <= dataout1_R1 WHEN dataout_sel(1) = '1' ELSE dataout1 ;
   dataout_int0 <= dataout0_R1 WHEN dataout_sel(0) = '1' ELSE dataout0 ;

   deskew_en <= op_sel(0);
   -- Instantiate deskew module that implements bit alignment algorithm
   deskew_module : deskew 
      PORT MAP (
         datain => datain,
         edgei => edgei_training,
         rst => rst,
         deskew_en => deskew_en,
         clkdiv => clkdiv,
         dlyce => dlyce_training,
         dlyinc => dlyinc_training,
         done => deskew_done);   
   
   bitslip_en <= op_sel(1);
   -- Instantiate the control module that implements the word alignment algorithm
   bitslip_control_module : bitslip_ctrl 
      PORT MAP (
         datain => datain,
         rst => rst,
         bitslip_en => bitslip_en,
         clkdiv => clkdiv,
         bitslip => bitslip_training,
         done => bitslip_done,
         error => bitslip_error);   
   
   
   -- Instantiate the control module that implements the real-time window monitoring algorithm
   monitor_ctrl_module : monitor_ctrl 
      PORT MAP (
         datain => datain,
         datain_monitor => datain_monitor,
         rst => rst,
         monitor_en => train_done_int,
         clkdiv => clkdiv,
         dlyce_window => dlyce_window,
         dlyinc_window => dlyinc_window,
         dlyce => dlyce_monitor,
         dlyinc => dlyinc_monitor,
         monitor_done => monitor_done);   
   

   -- Multi-purpose Counter   
   PROCESS(rst, clkdiv)
   BEGIN
      IF (rst = '1') THEN
         counter <= (others => '0');  
      ELSIF rising_edge(clkdiv) THEN
         IF (cnt_rst = '1') THEN
            counter <= (others => '0'); 
         ELSIF (counter(2) = '1') THEN
            counter <= (others => '0');
	    ELSE
            counter <= counter + X"01"; 
	    END IF;	  	             
      END IF;
   END PROCESS;

   -- Module that controls the operation select (bit align, word align, bitslip adjust   
   -- or window monitoring) and channel select.   
   PROCESS (rst, clkdiv)
   BEGIN
      IF (rst = '1') THEN
         op_sel <= "0000";    
         train_done_int <= '0';    
         chan_sel <= (OTHERS => '0');    
         cnt_rst <= '1';    
         dataout_sel <= (OTHERS => '0');
	 ELSIF rising_edge(clkdiv) THEN      
      IF (train_en = '1') THEN
         IF (op_sel_red_or = '0') THEN
            op_sel <= "0001";    
            train_done_int <= '0';    
            chan_sel <= X"0001";    
            cnt_rst <= '1';    
            dataout_sel <= (OTHERS => '0');    
         ELSIF (op_sel(0) = '1') THEN
            IF (chan_sel_red_or = '0') THEN
               op_sel <= STD_LOGIC_VECTOR(UNSIGNED(op_sel) SLL 1);    
               train_done_int <= '0';    
               chan_sel <= X"0001";    
               cnt_rst <= '1';    
               dataout_sel <= (OTHERS => '0');    
            ELSIF (deskew_done = '1') THEN
               op_sel <= op_sel;    
               train_done_int <= '0';    
               chan_sel <= STD_LOGIC_VECTOR(UNSIGNED(chan_sel) SLL 1);    
               cnt_rst <= '1';    
               dataout_sel <= (OTHERS => '0');    
            ELSE
               op_sel <= op_sel;    
               train_done_int <= '0';    
               chan_sel <= chan_sel;    
               cnt_rst <= '1';    
               dataout_sel <= (OTHERS => '0');    
            END IF;
         ELSIF (op_sel(1) = '1') THEN
            IF (chan_sel_red_or = '0') THEN
               op_sel <= STD_LOGIC_VECTOR(UNSIGNED(op_sel) SLL 1);    
               train_done_int <= '0';    
               chan_sel <= X"8000";    
               cnt_rst <= '0';    
               dataout_sel <= (OTHERS => '0');    
            ELSIF (bitslip_done = '1') THEN
               op_sel <= op_sel;    
               train_done_int <= '0';    
               chan_sel <= STD_LOGIC_VECTOR(UNSIGNED(chan_sel) SLL 1);    
               cnt_rst <= '1';    
               dataout_sel <= (OTHERS => '0');    
            ELSE
               op_sel <= op_sel;    
               train_done_int <= '0';    
               chan_sel <= chan_sel;    
               cnt_rst <= '1';    
               dataout_sel <= (OTHERS => '0');    
            END IF;
         ELSIF (op_sel(2) = '1') THEN
            IF (chan_sel_red_or = '0') THEN
               op_sel <= STD_LOGIC_VECTOR(UNSIGNED(op_sel) SLL 1);    
               train_done_int <= '1';    
               chan_sel <= X"0001";    
               cnt_rst <= '1';    
               IF (polarity(1) = '1') THEN
                  IF (polarity(0) = '1') THEN
                     dataout_sel <= dataout_sel_int;    
                  ELSE
                     dataout_sel <= NOT dataout_sel_int;    
                  END IF;
               ELSE
                  dataout_sel <= (OTHERS => '0');    
               END IF;
            -- If counter = 3, increment chan sel                        
            ELSIF ((counter(1) AND counter(0)) = '1') THEN
               op_sel <= op_sel;    
               train_done_int <= '0';    
               chan_sel <= STD_LOGIC_VECTOR(UNSIGNED(chan_sel) SRL 1);    
               cnt_rst <= '0';    
               dataout_sel <= dataout_sel;    
            ELSE
               op_sel <= op_sel;    
               train_done_int <= '0';    
               chan_sel <= chan_sel;    
               cnt_rst <= '0';    
               dataout_sel <= dataout_sel;    
            END IF;
         ELSIF (op_sel(3) = '1') THEN
            op_sel <= op_sel;    
            train_done_int <= '1';    
            cnt_rst <= '0';    
            dataout_sel <= dataout_sel;    
            IF (chan_sel_red_or = '0') THEN
               chan_sel <= X"0001";    
            ELSIF (monitor_done = '1') THEN
               chan_sel <= STD_LOGIC_VECTOR(UNSIGNED(chan_sel) SLL 1);    
            ELSE
               chan_sel <= chan_sel;    
            END IF;
         ELSE
            op_sel <= "0000";    
            train_done_int <= '0';    
            chan_sel <= (OTHERS => '0');    
            cnt_rst <= '1';    
            dataout_sel <= (OTHERS => '0');    

         END IF;
      ELSE -- !train_en            
         IF (op_sel(3) = '1') THEN
            op_sel <= op_sel;    
            train_done_int <= '1';    
            cnt_rst <= '0';    
            dataout_sel <= dataout_sel;    
            IF (chan_sel_red_or = '0') THEN
               chan_sel <= X"0001";    
            ELSIF (monitor_done = '1') THEN
               chan_sel <= STD_LOGIC_VECTOR(UNSIGNED(chan_sel) SLL 1);    
            ELSE
               chan_sel <= chan_sel;    
            END IF;
         ELSE
            op_sel <= "0000";    
            train_done_int <= '0';    
            chan_sel <= (OTHERS => '0');    
            cnt_rst <= '1';    
            dataout_sel <= (OTHERS => '0');    
         END IF;
      END IF;
      END IF;
   END PROCESS;

   -- Dataout Channel Select Mux
   muxout_d0d1 <= dataout1_R1 WHEN chan_sel(1) = '1' ELSE dataout0_R1 ;
   muxout_d2d3 <= dataout3_R1 WHEN chan_sel(3) = '1' ELSE dataout2_R1 ;
   muxout_d4d5 <= dataout5_R1 WHEN chan_sel(5) = '1' ELSE dataout4_R1 ;
   muxout_d6d7 <= dataout7_R1 WHEN chan_sel(7) = '1' ELSE dataout6_R1 ;
   muxout_d8d9 <= dataout9_R1 WHEN chan_sel(9) = '1' ELSE dataout8_R1 ;
   muxout_d10d11 <= dataout11_R1 WHEN chan_sel(11) = '1' ELSE dataout10_R1 ;
   muxout_d12d13 <= dataout13_R1 WHEN chan_sel(13) = '1' ELSE dataout12_R1 ;
   muxout_d14d15 <= dataout15_R1 WHEN chan_sel(15) = '1' ELSE dataout14_R1 ;

   muxout_d0_to_d3 <= muxout_d2d3 WHEN (chan_sel(2) OR chan_sel(3)) = '1' ELSE muxout_d0d1 ;
   muxout_d4_to_d7 <= muxout_d6d7 WHEN (chan_sel(6) OR chan_sel(7)) = '1' ELSE muxout_d4d5 ;
   muxout_d8_to_d11 <= muxout_d10d11 WHEN (chan_sel(10) OR chan_sel(11)) = '1' ELSE muxout_d8d9 ;
   muxout_d12_to_d15 <= muxout_d14d15 WHEN (chan_sel(14) OR chan_sel(15)) = '1' ELSE muxout_d12d13 ;

   -- Pipeline Registers inserted in this stage
   muxout_d0_to_d7 <= muxout_d4_to_d7_R1 WHEN (chan_sel(4) OR chan_sel(5) OR chan_sel(6) OR chan_sel(7)) = '1' ELSE muxout_d0_to_d3_R1 ;
   muxout_d8_to_d15 <= muxout_d12_to_d15_R1 WHEN (chan_sel(12) OR chan_sel(13) OR chan_sel(14) OR chan_sel(15)) = '1' ELSE muxout_d8_to_d11_R1 ;

   datain <= muxout_d8_to_d15 WHEN (chan_sel(8) OR chan_sel(9) OR chan_sel(10) OR chan_sel(11) OR chan_sel(12) OR chan_sel(13) OR chan_sel(14) OR chan_sel(15)) = '1' ELSE muxout_d0_to_d7 ;

   -- Window Monitor Channel Select Mux
   muxout_w0w1 <= windowout1_R1 WHEN chan_sel(1) = '1' ELSE windowout0_R1 ;
   muxout_w2w3 <= windowout3_R1 WHEN chan_sel(3) = '1' ELSE windowout2_R1 ;
   muxout_w4w5 <= windowout5_R1 WHEN chan_sel(5) = '1' ELSE windowout4_R1 ;
   muxout_w6w7 <= windowout7_R1 WHEN chan_sel(7) = '1' ELSE windowout6_R1 ;
   muxout_w8w9 <= windowout9_R1 WHEN chan_sel(9) = '1' ELSE windowout8_R1 ;
   muxout_w10w11 <= windowout11_R1 WHEN chan_sel(11) = '1' ELSE windowout10_R1 ;
   muxout_w12w13 <= windowout13_R1 WHEN chan_sel(13) = '1' ELSE windowout12_R1 ;
   muxout_w14w15 <= windowout15_R1 WHEN chan_sel(15) = '1' ELSE windowout14_R1 ;

   muxout_w0_to_w3 <= muxout_w2w3 WHEN (chan_sel(2) OR chan_sel(3)) = '1' ELSE muxout_w0w1 ;
   muxout_w4_to_w7 <= muxout_w6w7 WHEN (chan_sel(6) OR chan_sel(7)) = '1' ELSE muxout_w4w5 ;
   muxout_w8_to_w11 <= muxout_w10w11 WHEN (chan_sel(10) OR chan_sel(11)) = '1' ELSE muxout_w8w9 ;
   muxout_w12_to_w15 <= muxout_w14w15 WHEN (chan_sel(14) OR chan_sel(15)) = '1' ELSE muxout_w12w13 ;

   -- Pipeline Registers inserted in this stage
   muxout_w0_to_w7 <= muxout_w4_to_w7_R1 WHEN (chan_sel(4) OR chan_sel(5) OR chan_sel(6) OR chan_sel(7)) = '1' ELSE muxout_w0_to_w3_R1 ;
   muxout_w8_to_w15 <= muxout_w12_to_w15_R1 WHEN (chan_sel(12) OR chan_sel(13) OR chan_sel(14) OR chan_sel(15)) = '1' ELSE muxout_w8_to_w11_R1 ;

   datain_monitor <= muxout_w8_to_w15 WHEN (chan_sel(8) OR chan_sel(9) OR chan_sel(10) OR chan_sel(11) OR chan_sel(12) OR chan_sel(13) OR chan_sel(14) OR chan_sel(15)) = '1' ELSE muxout_w0_to_w7 ;

   -- Pipline Registers for Mux   
   PROCESS (rst, clkdiv)
   BEGIN
      IF (rst = '1') THEN
         muxout_d0_to_d3_R1 <= "0000";    
         muxout_d4_to_d7_R1 <= "0000";    
         muxout_d8_to_d11_R1 <= "0000";    
         muxout_d12_to_d15_R1 <= "0000";    
         muxout_w0_to_w3_R1 <= "0000";    
         muxout_w4_to_w7_R1 <= "0000";    
         muxout_w8_to_w11_R1 <= "0000";    
         muxout_w12_to_w15_R1 <= "0000";    
      ELSIF rising_edge(clkdiv) THEN
         muxout_d0_to_d3_R1 <= muxout_d0_to_d3;    
         muxout_d4_to_d7_R1 <= muxout_d4_to_d7;    
         muxout_d8_to_d11_R1 <= muxout_d8_to_d11;    
         muxout_d12_to_d15_R1 <= muxout_d12_to_d15;    
         muxout_w0_to_w3_R1 <= muxout_w0_to_w3;    
         muxout_w4_to_w7_R1 <= muxout_w4_to_w7;    
         muxout_w8_to_w11_R1 <= muxout_w8_to_w11;    
         muxout_w12_to_w15_R1 <= muxout_w12_to_w15;    
      END IF;
   END PROCESS;

   dlyce1 <= dlyce_monitor WHEN train_done_int = '1' ELSE dlyce_training ;
   dlyce_iserdes_window1 <= dlyce_window WHEN train_done_int = '1' ELSE dlyce_training ;
   dlyinc1 <= dlyinc_monitor WHEN train_done_int = '1' ELSE dlyinc_training ;
   dlyinc_iserdes_window1 <= dlyinc_window WHEN train_done_int = '1' ELSE dlyinc_training ;

END top_arch;
